# encoding: utf8
from pygubu import BuilderObject, register_widget, register_property
from pygubu.builder.ttkstdwidgets import TTKCombobox
from pygubu.widgets.combobox import Combobox


class ComboboxBuilder(TTKCombobox):
    OPTIONS_SPECIFIC = tuple(set(TTKCombobox.OPTIONS_SPECIFIC) - set(('state',)))
    OPTIONS_CUSTOM = TTKCombobox.OPTIONS_CUSTOM + ('keyvariable',)
    properties = (TTKCombobox.OPTIONS_STANDARD + TTKCombobox.OPTIONS_SPECIFIC +
                  OPTIONS_CUSTOM)
    tkvar_properties = TTKCombobox.tkvar_properties + ('keyvariable',)
    class_ = Combobox

register_widget('pygubu.builder.widgets.combobox', ComboboxBuilder, 'Combobox',
                ('ttk', 'Pygubu Widgets'))

props = {
    'keyvariable': {
        'editor': 'tkvarentry'
        },
    'state': {
        'editor': 'choice',
        'pygubu.builder.widgets.combobox': {
            'params': {
                'values': ('', 'normal', 'disabled'),
                'state': 'readonly'}},
        }
    }

for p in props:
    register_property(p, props[p])



# encoding: utf8
from __future__ import unicode_literals
import logging
import itertools
import json
from collections import namedtuple, defaultdict

try:
    import tkinter as tk
except:
    import Tkinter as tk

__all__ = [
    'BuilderObject', 'EntryBaseBO', 'PanedWindowBO',
    'PanedWindowPaneBO', 'WidgetDescription', 'CLASS_MAP', 'CB_TYPES',
    'CUSTOM_PROPERTIES', 'register_widget', 'register_property']

logger = logging.getLogger(__name__)

#
# Utility functions
#
zip_longest = getattr(itertools, 'zip_longest', None)
if zip_longest is None:
    def zip_longest(*args, **kw):
        # zip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
        fillvalue = kw.get('fillvalue', None)
        iterators = [iter(it) for it in args]
        num_active = len(iterators)
        if not num_active:
            return
        while True:
            values = []
            for i, it in enumerate(iterators):
                try:
                    value = next(it)
                except StopIteration:
                    num_active -= 1
                    if not num_active:
                        return
                    iterators[i] = itertools.repeat(fillvalue)
                    value = fillvalue
                values.append(value)
            yield tuple(values)
    itertools.zip_longest = zip_longest


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


# Python 2 issue:
try:
    isinstance(basestring, type)
except:
    basestring = str

#
# BuilderObject
#

WidgetDescription = namedtuple('WidgetDescription',
                              ['classname', 'builder', 'label', 'tags'])

CLASS_MAP = {}


def register_widget(classname, builder, label=None, tags=None):
    if label is None:
        label = classname
    if tags is None:
        tags = tuple()

    CLASS_MAP[classname] = WidgetDescription(classname, builder, label, tags)


CUSTOM_PROPERTIES = {}


def register_property(name, description):
    if name in CUSTOM_PROPERTIES:
        CUSTOM_PROPERTIES[name].update(description)
        logger.debug('Updating registered property %s', name)
    else:
        CUSTOM_PROPERTIES[name] = description
        logger.debug('Registered property %s', name)


class CB_TYPES:
    '''Callback types'''
    SIMPLE = 'simple'
    WITH_WID = 'with_wid'
    ENTRY_VALIDATE = 'entry_validate'
    SCROLL = 'scroll'
    SCROLLSET = 'scrollset'
    SCALE = 'scale'
    BIND_EVENT = 'bind_event'

#
# Base class
#
class BuilderObject(object):
    """Base class for Widgets created with Builder"""

    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = tuple()
    OPTIONS_CUSTOM = tuple()
    class_ = None
    container = False
    allowed_parents = None
    allowed_children = None
    maxchildren = None
    properties = tuple()
    ro_properties = tuple()
    layout_required = True
    command_properties = tuple()
    allow_bindings = True
    tkvar_properties = ('listvariable', 'textvariable', 'variable')
    tkimage_properties = ('image', 'selectimage', 'iconphoto')
    tkfont_properties = ('font',)
    virtual_events = tuple()

    @classmethod
    def factory(cls, builder, wdata):
        clsobj = cls(builder, wdata)
        wdata.layout_required = clsobj.layout_required
        return clsobj

    def __init__(self, builder, wmeta):
        super(BuilderObject, self).__init__()
        self.widget = None
        self.builder = builder
        self.wmeta = wmeta
        self._code_identifier = None

    def realize(self, parent):
        args = self._get_init_args()
        master = parent.get_child_master()
        self.widget = self.class_(master, **args)
        return self.widget

    def _get_init_args(self):
        """Creates dict with properties marked as readonly"""

        args = {}
        for rop in self.ro_properties:
            if rop in self.wmeta.properties:
                pvalue = self._process_property_value(
                    rop, self.wmeta.properties[rop])
                args[rop] = pvalue
        return args

    def configure(self, target=None):
        if target is None:
            target = self.widget
        for pname, value in self.wmeta.properties.items():
            if (pname not in self.ro_properties and
                pname not in self.command_properties):
                pvalue = self._process_property_value(pname, value)
                self._set_property(target, pname, pvalue)
                
    def _process_property_value(self, pname, value):
        return value

    def _set_property(self, target_widget, pname, value):
        if pname not in self.__class__.properties:
            msg = "Attempt to set an unknown property '%s' on class '%s'"
            logger.warning(msg, pname, repr(self.class_))
        else:
            propvalue = value
            if pname in self.tkvar_properties:
                propvalue = self.builder.create_variable(value)
                if 'text' in self.wmeta.properties and pname == 'textvariable':
                    propvalue.set(self.wmeta.properties['text'])
                elif 'value' in self.wmeta.properties and pname == 'variable':
                    propvalue.set(self.wmeta.properties['value'])
            elif pname in self.tkimage_properties:
                propvalue = self.builder.get_image(value)
            elif pname == 'takefocus':
                if value:
                    propvalue = tk.getboolean(value)

            try:
                logger.debug('setting property %s = %s', pname, propvalue)
                target_widget[pname] = propvalue
            except tk.TclError as e:
                msg = "Failed to set property '{0}' on class '{1}'. TclError:"
                logger.error(msg, pname, repr(self.class_))
                logger.exception(e)

    def layout(self, target=None, configure_gridrc=True):
        if self.layout_required:
            if target is None:
                target = self.widget
    
            # Check manager
            manager = self.wmeta.manager
            logger.debug('Applying %s layout to %s', manager, self.wmeta.identifier)
            if manager == 'grid':
                self._grid_layout(target)
            elif manager == 'pack':
                self._pack_layout(target)
            elif manager == 'place':
                self._place_layout(target)
            else:
                msg = 'Invalid layout manager: {0}'.format(manager)
                raise Exception(msg)
        if configure_gridrc:
            logger.debug('Configurying grid-rc')
            parent = target.nametowidget(target.winfo_parent())
            self._gridrc_config(parent)

    def _pack_layout(self, target):
        properties = dict(self.wmeta.layout_properties)
        propagate = properties.pop('propagate', 'true')
        # Do pack
        target.pack(**properties)
        if propagate.lower() != 'true':
            target.pack_propagate(0)
    
    def _place_layout(self, target):
        # Do place
        target.place(**self.wmeta.layout_properties)
    
    def _grid_layout(self, target):
        properties = dict(self.wmeta.layout_properties)
        propagate = properties.pop('propagate', 'true')
        propagate = propagate.lower()
        target.grid(**properties)
        if propagate != 'true':
            target.grid_propagate(0)

    def _gridrc_config(self, target):
        # configure grid row/col properties:
        for type_, num, pname, value in self.wmeta.gridrc_properties:
            if type_ == 'row':
                target.rowconfigure(num, **{pname: value})
            else:
                target.columnconfigure(num, **{pname: value})

    def get_child_master(self):
        return self.widget

    def add_child(self, bobject):
        pass

    def _create_callback(self, cmd, command):
        callback = command
        cmd_type = cmd['cbtype']
        if cmd_type == CB_TYPES.WITH_WID:
            def widget_callback(button_id=self.wmeta.identifier):
                command(button_id)
            callback = widget_callback
        if cmd_type == CB_TYPES.ENTRY_VALIDATE:
            args = cmd['args']
            if args:
                args = args.split(' ')
                callback = (self.widget.register(command),) + tuple(args)
            else:
                callback = self.widget.register(command)
        return callback

    def _connect_command(self, cmd_pname, callback):
        prop = {cmd_pname: callback}
        self.widget.configure(**prop)

    def connect_commands(self, cmd_bag):
        notconnected = []
        commands = {}
        for cmd_pname in self.command_properties:
            cmd = self.wmeta.properties.get(cmd_pname, None)
            if cmd is not None:
                cmd = json.loads(cmd)
                cmd_name = (cmd['value']).strip()
                if cmd_name:
                    commands[cmd_pname]= cmd
                else:
                    msg = "%s: invalid value for property '%s'."
                    logger.warning(msg, self.wmeta.identifier, cmd_pname)
        
        if isinstance(cmd_bag, dict):
            for cmd_pname, cmd in commands.items():
                cmd_name = cmd['value']
                if cmd_name in cmd_bag:
                    callback = self._create_callback(cmd, cmd_bag[cmd_name])
                    self._connect_command(cmd_pname, callback)
                else:
                    notconnected.append(cmd_name)
        else:
            for cmd_pname, cmd in commands.items():
                cmd_name = cmd['value']
                if hasattr(cmd_bag, cmd_name):
                    callback = self._create_callback(
                        cmd, getattr(cmd_bag, cmd_name))
                    self._connect_command(cmd_pname, callback)
                else:
                    notconnected.append(cmd_name)
        if notconnected:
            return notconnected
        else:
            return None

    def connect_bindings(self, cb_bag):
        notconnected = []

        if isinstance(cb_bag, dict):
            for bind in self.wmeta.bindings:
                cb_name = bind.handler
                if cb_name in cb_bag:
                    callback = cb_bag[cb_name]
                    self.widget.bind(bind.sequence, callback, add=bind.add)
                else:
                    notconnected.append(cb_name)
        else:
            for bind in self.wmeta.bindings:
                cb_name = bind.handler
                if hasattr(cb_bag, cb_name):
                    callback = getattr(cb_bag, cb_name)
                    self.widget.bind(bind.sequence, callback, add=bind.add)
                else:
                    notconnected.append(cb_name)
        if notconnected:
            return notconnected
        else:
            return None

    #
    # Code generation methods
    #
    def code_realize(self, boparent, code_identifier=None):
        if code_identifier is not None:
            self._code_identifier = code_identifier
        lines = []
        master = boparent.code_child_master()
        init_args = self._get_init_args()
        bag = []
        for pname, value in init_args.items():
            s = "{0}='{1}'".format(pname, value)
            bag.append(s)
        kwargs = ''
        if bag:
            kwargs = ', {0}'.format(', '.join(bag))
        s = "{0} = {1}({2}{3})".format(self.code_identifier(),
                                       self._code_class_name(), master, kwargs)
        lines.append(s)
        return lines
    
    def code_identifier(self):
        if self._code_identifier is None:
            return self.wmeta.identifier
        return self._code_identifier
    
    def code_child_master(self):
        return self.code_identifier()
    
    def code_child_add(self, childid):
        return []
    
    def code_configure(self, targetid=None):
        if targetid is None:
            targetid = self.code_identifier()
        code_bag, kwproperties, complex_properties = \
            self._code_process_properties(self.wmeta.properties, targetid)
        lines = []
        prop_stmt = "{0}.configure({1})"
        arg_stmt = "{0}={1}"
        for g in grouper(sorted(kwproperties), 4):
            args_bag = []
            for p in g:
                if p is not None:
                    args_bag.append(arg_stmt.format(p, code_bag[p]))
            args = ', '.join(args_bag)
            line = prop_stmt.format(targetid, args)
            lines.append(line)
        for pname in complex_properties:
            lines.extend(code_bag[pname])
        
        return lines
    
    def code_layout(self, targetid=None):
        if targetid is None:
            targetid = self.code_identifier()
        if self.layout_required:
            lines = []
            layout_stmt = "{0}.{1}({2})"
            arg_stmt = "{0}='{1}'"
            layout = self.wmeta.layout_properties
            args_bag = []
            for p, v in sorted(layout.items()):
                if p not in ('propagate', ):
                    args_bag.append(arg_stmt.format(p, v))
            args = ', '.join(args_bag)
            
            manager = self.wmeta.manager
            line = layout_stmt.format(targetid, manager, args)
            lines.append(line)
            
            pvalue = str(layout.get('propagate', '')).lower()
            if 'propagate' in layout and  pvalue == 'false':
                line = '{0}.{1}_propagate({2})'
                line = line.format(targetid, manager, 0)
                lines.append(line)
            
            lrow_stmt = "{0}.rowconfigure('{1}', {2})"
            lcol_stmt = "{0}.columnconfigure('{1}', {2})"
            rowbag = defaultdict(list)
            colbag = defaultdict(list)
            for type_, num, pname, value in self.wmeta.gridrc_properties:
                arg = "{0}='{1}'".format(pname, value)
                if type_ == 'row':
                    rowbag[num].append(arg)
                else:
                    colbag[num].append(arg)
            for k, bag in rowbag.items():
                args = ', '.join(bag)
                line = lrow_stmt.format(targetid, k, args)
                lines.append(line)
            for k, bag in colbag.items():
                args = ', '.join(bag)
                line = lcol_stmt.format(targetid, k, args)
                lines.append(line)
            return lines
        else:
            return []
    
    def _code_class_name(self):
        cname = None
        # try here ?
        cname = self.builder.code_classname_for(self)
        if cname is None:
            if self.class_ is not None:
                prefix = self.class_.__module__
                name = self.class_.__name__
                cname = '{0}.{1}'.format(prefix, name)
            else:
                cname = 'ClassNameNotDefined'
        return cname
    
    def _code_process_properties(self, properties, targetid):
        code_bag = {}
        for pname, value in properties.items():
            if (pname not in self.ro_properties and
                pname not in self.command_properties):
                self._code_set_property(targetid, pname, value, code_bag)
        
        # properties
        # determine kw properties or complex properties
        kwproperties = []
        complex_properties = []
        for pname, value in code_bag.items():
            if isinstance(value, str) or isinstance(value, basestring):
                kwproperties.append(pname)
            else:
                complex_properties.append(pname)
        
        return (code_bag, kwproperties, complex_properties)
    
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname in self.OPTIONS_CUSTOM:
            line = "# TODO - {0}: code for custom option '{1}' not implemented."
            line = line.format(targetid, pname)
            code_bag[pname] = [line]
        else:
            propvalue = "'{}'".format(value)
            if pname in self.tkvar_properties:
                varvalue = None
                if 'text' in self.wmeta.properties and pname == 'textvariable':
                    varvalue = self.wmeta.properties['text']
                elif 'value' in self.wmeta.properties and pname == 'variable':
                    varvalue = self.wmeta.properties['value']
                propvalue = self.builder.code_create_variable(value, varvalue)
            elif pname in self.command_properties:
                cmd_name = value.strip()
                callback = self.builder.code_create_callback(
                    targetid, cmd_name, CB_TYPES.SIMPLE)
                propvalue = callback
            elif pname in self.tkimage_properties:
                propvalue = self.builder.code_create_image(value)
            elif pname == 'takefocus':
                propvalue = str(tk.getboolean(value))
            code_bag[pname] = propvalue
    
    def code_connect_commands(self):
        commands = {}
        for cmd_pname in self.command_properties:
            cmd = self.wmeta.properties.get(cmd_pname, None)
            if cmd is not None:
                #print('on_code_connect_commands', cmd)
                cmd = json.loads(cmd)
                cmd_name = (cmd['value']).strip()
                if cmd_name:
                    commands[cmd_pname]= cmd
                else:
                    msg = "%s: invalid callback name for property '%s'."
                    logger.warning(msg, self.wmeta.identifier, cmd_pname)
        lines = []
        for cmd_pname, cmd in commands.items():
            callback = self._code_define_callback(cmd_pname, cmd)
            cmd_code = self._code_connect_command(cmd_pname, callback)
            if cmd_code:
                lines.extend(cmd_code)
        return lines
    
    def _code_define_callback(self, cmd_pname, cmd):
        cmdname = cmd['value']
        cmdtype = cmd['cbtype']
        #print('_code_define_callback', cmd)
        wid = self.code_identifier()
        args = None
        if cmdtype == CB_TYPES.WITH_WID:
            args = ('widget_id', )
        if cmdtype == CB_TYPES.SCALE:
            args = ('scale_value', )
        if cmdtype == CB_TYPES.SCROLLSET:
            args = ('first', 'last')
        if cmdtype == CB_TYPES.SCROLL:
            args = ('mode', 'value', 'units')
        return self.builder.code_create_callback(wid, cmdname, cmdtype, args)
    
    def _code_connect_command(self, cmd, cbname):
        target = self.code_identifier()
        line = '{0}.configure({1}={2})'.format(target, cmd, cbname)
        return (line, )
    
    def code_connect_bindings(self):
        lines = []
        target = self.code_identifier()
        for bind in self.wmeta.bindings:
            cb_name = self.builder.code_create_callback(
                target, bind.handler, CB_TYPES.BIND_EVENT)
            add_arg = '+' if bind.add else ''
            line = "{0}.bind('{1}', {2}, add='{3}')"
            line = line.format(target, bind.sequence, cb_name, add_arg)
            lines.append(line)
        return lines


#
# Base clases for some widgets
#
class EntryBaseBO(BuilderObject):
    """Base class for tk.Entry and ttk.Entry builder objects"""
    def _set_property(self, target_widget, pname, value):
        if pname == 'text':
            wstate = str(target_widget['state'])
            if wstate != 'normal':
                # change state temporarily
                target_widget['state'] = 'normal'
            target_widget.delete('0', tk.END)
            target_widget.insert('0', value)
            target_widget['state'] = wstate
        else:
            super(EntryBaseBO, self)._set_property(target_widget, pname, value)
    
    #
    # Code generation methods
    #
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == 'text':
            lines = [
                "_text_ = '''{0}'''".format(value),
                "{0}.delete('0', 'end')".format(targetid),
                "{0}.insert('0', _text_)".format(targetid),
                ]
            if 'state' in self.wmeta.properties:
                state_value = self.wmeta.properties['state']
                if state_value != 'normal':
                    line = "{0}['state'] = 'normal'".format(targetid)
                    lines.insert(1, line)
                    line = "{0}['state'] = '{1}'".format(targetid, state_value)
                    lines.append(line)
            code_bag[pname] = lines
        else:
            super(EntryBaseBO, self)._code_set_property(targetid, pname,
                                                        value, code_bag)


class PanedWindowBO(BuilderObject):
    """Base class for tk.PanedWindow and ttk.Panedwindow builder objects"""
    class_ = None
    container = True
    properties = []
    ro_properties = ('orient', )

    def realize(self, parent):
        master = parent.get_child_master()
        args = self._get_init_args()
        if 'orient' not in args:
            args['orient'] = 'vertical'
        self.widget = self.class_(master, **args)
        return self.widget


#
# Base clases for some widget Helpers
#
class PanedWindowPaneBO(BuilderObject):
    class_ = None
    container = True
    properties = []
    layout_required = False
    allow_bindings = False

    def realize(self, parent):
        self.widget = parent.get_child_master()
        return self.widget

    def configure(self):
        pass

    def layout(self):
        pass

    def add_child(self, bobject):
        self.widget.add(bobject.widget, **self.wmeta.properties)
    
    #
    # code generation functions
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        return tuple()
    
    def code_configure(self, targetid=None):
        return tuple()
    
    def code_layout(self, targetid=None):
        return tuple()
    
    def code_child_add(self, childid):
        config = []
        masterid = self.code_child_master()
        for pname, val in self.wmeta.properties.items():
            line = "{}='{}'".format(pname, val)
            config.append(line)
        kw = ''
        lines = []
        if config:
            kw = ', {0}'.format(', '.join(config))
        line = '{0}.add({1}{2})'.format(masterid, childid, kw)
        lines.append(line)
        return lines



# encoding: utf8
from __future__ import unicode_literals

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from pygubu.widgets.tkscrollbarhelper import ScrollbarHelperBase


class TTKScrollbarHelperFactory(type):
    def __new__(cls, clsname, superclasses, attrs):
        return type.__new__(cls, str(clsname), superclasses, attrs)


ScrollbarHelper = TTKScrollbarHelperFactory('ScrollbarHelper',
                                       (ScrollbarHelperBase, ttk.Frame, object),
                                       {'_framecls': ttk.Frame,
                                        '_sbarcls': ttk.Scrollbar})



# encoding: UTF-8
from __future__ import unicode_literals

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

img_down = '''\
R0lGODlhEAAQAIAAAAAAAAAAACH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEKAAEALAAAAAAQABAA
AAIXjI+py+0P4wK0WprunRo0/VgRJpXmyRQAOw==
'''

img_right = '''\
R0lGODlhEAAQAIAAAAAAAAAAACH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEKAAEALAAAAAAQABAA
AAIdjI+pywGtwINHTmpvy3rxnnwQh1mUI52o6rCuWgAAOw==
'''


class AccordionFrame(ttk.Frame):
    """ An accordion like widget.
    Usage:
        acframe = AccordionFrame(master)
        acframe.grid()
        g = acframe.add_group('g1', 'Group1')
        label = ttk.Label(g, text='Label on group1')
        label.grid()
    """
    IMAGES = None

    def __init__(self, master=None, **kw):
        ttk.Frame.__init__(self, master, **kw)
        self.__images = None
        self.__groups = {}
        self.columnconfigure(0, weight=1)

        if AccordionFrame.IMAGES is None:
            AccordionFrame.IMAGES = [
                tk.PhotoImage(data=img_down),
                tk.PhotoImage(data=img_right),]
        self.__images = AccordionFrame.IMAGES



    def add_group(self, gid, label=None, expanded=True):

        glabel = label
        if label is None:
            glabel = str(gid)

        #button creation
        btn = ttk.Button(self, text=glabel, style='Toolbutton',
            image=self.__images[0], compound='left')
        btn.grid(sticky=tk.EW)
        btn.dd_show = True
        btn.configure(command=lambda:self.__button_clicked(gid))

        #frame creation
        frame = ttk.Frame(self, width=100, height=100)
        frame.grid(sticky=tk.NSEW)
        frame.rowconfigure(0, weight=1)
        frame.columnconfigure(0, weight=1)

        if not expanded:
            self.after_idle(lambda: self.__button_clicked(gid))

        #store button, and frame
        self.__groups[gid] = (btn, frame)

        return frame


    def get_group(self, gid):
        return self.__groups[gid][1]


    def group_toogle(self, gid):
        self.__button_clicked(gid)


    def __button_clicked(self, gid):
        btn, frame = self.__groups[gid]
        if btn.dd_show == True:
            btn.dd_show = False
            btn.configure(image=self.__images[1])
            frame.grid_remove()
        else:
            btn.dd_show = True
            btn.configure(image=self.__images[0])
            frame.grid()
        self.event_generate('<<AccordionGroupToggle>>')


    def set_images(self, img_open, img_close):
        if self.__images == AccordionFrame.IMAGES:
            self.__images = [
                tk.PhotoImage(file=img_open),
                tk.PhotoImage(file=img_close),]
        else:
            self.__images[0].configure(file=img_open)
            self.__images[1].configure(file=img_close)


    @classmethod
    def set_gimages(cls, img_open, img_close):
        if cls.IMAGES is None:
            cls.IMAGES = [
                tk.PhotoImage(file=img_open),
                tk.PhotoImage(file=img_close),]
        else:
            cls.IMAGES[0].configure(file=img_open)
            cls.IMAGES[1].configure(file=img_close)


if __name__ == '__main__':
    root = tk.Tk()

    app = AccordionFrame(root)
    app.grid(sticky=tk.NSEW)

    top = app.winfo_toplevel()
    top.rowconfigure(0, weight=1)
    top.columnconfigure(0, weight=1)

    g = app.add_group('g1', 'Tk widgets')
    l = tk.Label(g, text="Label1")
    l.grid()
    l = tk.Label(g, text="Label2")
    l.grid()
    g = app.add_group('g2', 'Ttk widgets')
    ##
    app = AccordionFrame(g)
    app.grid(sticky='nsew', padx="5 0")
    g = app.add_group('g1', 'Containers')
    l = tk.Label(g, text="Label1")
    l.grid()
    l = tk.Label(g, text="Label2")
    l.grid()
    g = app.add_group('g2', 'Control')
    l = tk.Label(g, text="Label3")
    l.grid()
    l = tk.Label(g, text="Label4")
    l.grid()

    tk.mainloop()



from __future__ import unicode_literals

__all__ = ['remove_binding', 'ApplicationLevelBindManager']

import platform
import logging


logger = logging.getLogger(__name__)

def bindings(widget, seq):
    return [x for x in widget.bind(seq).splitlines() if x.strip()]

def _funcid(binding):
    return binding.split()[1][3:]

def remove_binding(widget, seq, index=None, funcid=None):
    b = bindings(widget, seq)

    if not index is None:
        try:
            binding = b[index]
            widget.unbind(seq, _funcid(binding))
            b.remove(binding)
        except IndexError:
            logger.info('Binding #%d not defined.', index)
            return

    elif funcid:
        binding = None
        for x in b:
            if _funcid(x) == funcid:
                binding = x
                b.remove(binding)
                widget.unbind(seq, funcid)
                break
        if not binding:
            logger.info( 'Binding "%s" not defined.', funcid)
            return
    else:
        raise ValueError('No index or function id defined.')

    for x in b:
        widget.bind(seq, '+'+x, 1)


class ApplicationLevelBindManager(object):
    # Mouse wheel support
    mw_active_area = None
    mw_initialized = False

    @staticmethod
    def on_mousewheel(event):
        if ApplicationLevelBindManager.mw_active_area:
            ApplicationLevelBindManager.mw_active_area.on_mousewheel(event)
    
    @staticmethod
    def mousewheel_bind(widget):
        ApplicationLevelBindManager.mw_active_area = widget
    
    @staticmethod
    def mousewheel_unbind():
        ApplicationLevelBindManager.mw_active_area = None

    @staticmethod
    def init_mousewheel_binding(master):
        if ApplicationLevelBindManager.mw_initialized == False:
            _os = platform.system()
            if _os in ('Linux', 'OpenBSD', 'FreeBSD'):
                master.bind_all('<4>', ApplicationLevelBindManager.on_mousewheel,  add='+')
                master.bind_all('<5>', ApplicationLevelBindManager.on_mousewheel,  add='+')
            else:
                # Windows and MacOS
                master.bind_all("<MouseWheel>", ApplicationLevelBindManager.on_mousewheel,  add='+')
            ApplicationLevelBindManager.mw_initialized = True

    @staticmethod
    def make_onmousewheel_cb(widget, orient, factor = 1):
        """Create a callback to manage mousewheel events
        
        orient: string (posible values: ('x', 'y'))
        widget: widget that implement tk xview and yview methods
        """
        _os = platform.system()
        view_command = getattr(widget, orient+'view')
        if _os in ('Linux', 'OpenBSD', 'FreeBSD'):
            def on_mousewheel(event):
                if event.num == 4:
                    view_command('scroll', (-1)*factor, 'units')
                elif event.num == 5:
                    view_command('scroll', factor, 'units') 
        
        elif _os == 'Windows':
            def on_mousewheel(event):        
                view_command('scroll', (-1)*int((event.delta/120)*factor), 'units') 
        
        elif _os == 'Darwin':
            def on_mousewheel(event):        
                view_command('scroll', event.delta, 'units')
        else:
            # FIXME: unknown platform scroll method
            def on_mousewheel(event):
                pass
        
        return on_mousewheel



# encoding: utf8
__all__ = ['ToolTip']

try:
    import tkinter as tk
    from tkinter import ttk
except:
    import Tkinter as tk
    import ttk


class ToolTip(object):

    def __init__(self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0
    
    def inside_wbbox(self, rx , ry):
        bbox = self._calc_bbox(self.widget, True)
        inside = False
        if (bbox[0] <= rx <= bbox[2]) and (bbox[1] <= ry <= bbox[3]):
            inside = True
        return inside
        
    def _calc_bbox(self, widget, screen=False):
        rx = widget.winfo_x()
        ry = widget.winfo_y()
        if screen:
            rx = widget.winfo_rootx()
            ry = widget.winfo_rooty()
        x2 = rx + widget.winfo_width()
        y2 = ry + widget.winfo_height()
        return (rx, ry, x2, y2)
    
    def _calc_final_pos(self):
        rx, ry, rcx, rcy = self._calc_bbox(self.widget, True)
        w = rcx-rx
        h = rcy-ry
        sh = self.widget.winfo_screenheight() - 10
        sw = self.widget.winfo_screenwidth() - 10
        x = y = 0
        for region in ('bottom', 'right', 'top', 'left'):
            if region == 'bottom':
                x = rx + int(w//2 * 0.2)
                y = rcy + int(h//2 * 0.1)
            elif region == 'right':
                x = rcx + int(w//2 * 0.2)
                y = ry + int(h//2 * 0.1)
            elif region == 'top':
                x = rx - int(w//2 * 0.4)
                y = ry - int(h//2 * 0.4)
            elif region == 'left':
                x = rx
                y = ry - 20
            if x < sw and y < sh:
                break
        return (x, y)

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y = self._calc_final_pos()
        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+{0}+{1}".format(x, y))
        try:
            # For Mac OS
            tw.tk.call("::tk::unsupported::MacWindowStyle",
                       "style", tw._w,
                       "help", "noActivates")
        except tk.TclError:
            pass
        label = tk.Label(tw, text=self.text, justify=tk.LEFT,
                      background="#ffffe0", foreground="black",
                      relief=tk.SOLID, borderwidth=1,
                      font=("tahoma", "9", "normal"))
        label.pack(ipadx=2)

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()

def create(widget, text):
    toolTip = ToolTip(widget)
    def enter(event):
        toolTip.showtip(text)
    def leave(event):
        toolTip.hidetip()
    widget.bind('<Enter>', enter)
    widget.bind('<Leave>', leave)


if __name__ == '__main__':
    root = tk.Tk()
    for idx in range(0, 2):
        b = tk.Button(root, text='A button')
        b.grid()
        create(b, 'A tooltip !!')
    root.mainloop()



# encoding: utf8
from pygubu import BuilderObject, register_widget, register_property
from pygubu.builder.ttkstdwidgets import TTKFrame
from pygubu.widgets.calendarframe import CalendarFrame


class CalendarFrameBuilder(BuilderObject):
    class_ = CalendarFrame
    OPTIONS_STANDARD = TTKFrame.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TTKFrame.OPTIONS_SPECIFIC
    OPTIONS_CUSTOM = ('firstweekday', 'year', 'month',
                      'calendarfg', 'calendarbg', 'headerfg', 'headerbg',
                      'selectbg', 'selectfg', 'state', 'markbg', 'markfg')
    ro_properties = TTKFrame.ro_properties
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    virtual_events = ('<<CalendarFrameDateSelected>>',)

register_widget('pygubu.builder.widgets.calendarframe', CalendarFrameBuilder,
                'CalendarFrame', ('ttk', 'Pygubu Widgets'))

props = {
    'state': {
        'editor': 'choice',
        'pygubu.builder.widgets.calendarframe': {
            'params': {
                'values': ('', 'normal', 'disabled'),
                'state': 'readonly'}},
        },
    'firstweekday': {
        'editor': 'choice',
        'params': {
            'values': ('0', '6'), 'state': 'readonly'},
        'default': '6'
        },
    'year': {
        'editor': 'entry'
        },
    'month': {
        'editor': 'choice',
        'params': {
            'values': ('1', '2', '3', '4', '5', '6', '7', '8', '9', '10',
                       '11', '12'), 'state': 'readonly'},
        'default': '1'
        },
# Better to change locale by code.
#    'locale': {
#        'editor': 'entry'
#        },
    'calendarfg': {
        'editor': 'colorentry'
        },
    'calendarbg': {
        'editor': 'colorentry'
        },
    'headerfg': {
        'editor': 'colorentry'
        },
    'headerbg': {
        'editor': 'colorentry'
        },
    'selectbg': {
        'editor': 'colorentry'
        },
    'selectfg': {
        'editor': 'colorentry'
        },
    'markbg': {
        'editor': 'colorentry'
        },
    'markfg': {
        'editor': 'colorentry'
        }
    }

for p in props:
    register_property(p, props[p])




# encoding: utf8
from __future__ import unicode_literals
import functools

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk


class EditableTreeview(ttk.Treeview):
    """A simple editable treeview

    It uses the following events from Treeview:
        <<TreviewSelect>>
        <4>
        <5>
        <KeyRelease>
        <Home>
        <End>
        <Configure>
        <Button-1>
        <ButtonRelease-1>
        <Motion>
    If you need them use add=True when calling bind method.

    It Generates two virtual events:
        <<TreeviewInplaceEdit>>
        <<TreeviewCellEdited>>
    The first is used to configure cell editors.
    The second is called after a cell was changed.
    You can know wich cell is being configured or edited, using:
        get_event_info()
    """
    def __init__(self, master=None, **kw):
        ttk.Treeview.__init__(self, master, **kw)

        self._curfocus = None
        self._inplace_widgets = {}
        self._inplace_widgets_show = {}
        self._inplace_vars = {}
        self._header_clicked = False
        self._header_dragged = False

        self.bind('<<TreeviewSelect>>', self.__check_focus)
        #Wheel events?
        self.bind('<4>', lambda e: self.after_idle(self.__updateWnds))
        self.bind('<5>', lambda e: self.after_idle(self.__updateWnds))
        #self.bind('<ButtonRelease-1>', self.__check_focus)
        self.bind('<KeyRelease>', self.__check_focus)
        self.bind('<Home>', functools.partial(self.__on_key_press, 'Home'))
        self.bind('<End>', functools.partial(self.__on_key_press, 'End'))
        self.bind('<Button-1>', self.__on_button1)
        self.bind('<ButtonRelease-1>', self.__on_button1_release)
        self.bind('<Motion>', self.__on_mouse_motion)
        self.bind('<Configure>',
            lambda e: self.after_idle(self.__updateWnds))


    def __on_button1(self, event):
        r = event.widget.identify_region(event.x, event.y)
        if r in ('separator', 'header'):
            self._header_clicked = True

    def __on_mouse_motion(self, event):
        if self._header_clicked:
            self._header_dragged = True

    def __on_button1_release(self, event):
        if self._header_dragged:
            self.after_idle(self.__updateWnds)
        self._header_clicked = False
        self._header_dragged = False

    def __on_key_press(self, key, event):
        if key == 'Home':
            self.selection_set("")
            self.focus(self.get_children()[0])
        if key == 'End':
            self.selection_set("")
            self.focus(self.get_children()[-1])

    def delete(self, *items):
        self.after_idle(self.__updateWnds)
        ttk.Treeview.delete(self, *items)

    def yview(self, *args):
        """Update inplace widgets position when doing vertical scroll"""
        self.after_idle(self.__updateWnds)
        ttk.Treeview.yview(self, *args)

    def yview_scroll(self, number, what):
        self.after_idle(self.__updateWnds)
        ttk.Treeview.yview_scroll(self, number, what)

    def yview_moveto(self, fraction):
        self.after_idle(self.__updateWnds)
        ttk.Treeview.yview_moveto(self, fraction)

    def xview(self, *args):
        """Update inplace widgets position when doing horizontal scroll"""
        self.after_idle(self.__updateWnds)
        ttk.Treeview.xview(self, *args)

    def xview_scroll(self, number, what):
        self.after_idle(self.__updateWnds)
        ttk.Treeview.xview_scroll(self, number, what)

    def xview_moveto(self, fraction):
        self.after_idle(self.__updateWnds)
        ttk.Treeview.xview_moveto(self, fraction)

    def __check_focus(self, event):
        """Checks if the focus has changed"""
        #print('Event:', event.type, event.x, event.y)
        changed = False
        if not self._curfocus:
            changed = True
        elif self._curfocus != self.focus():
            self.__clear_inplace_widgets()
            changed = True
        newfocus = self.focus()
        if changed:
            if newfocus:
                #print('Focus changed to:', newfocus)
                self._curfocus= newfocus
                self.__focus(newfocus)
            self.__updateWnds()

    def __focus(self, item):
        """Called when focus item has changed"""
        cols = self.__get_display_columns()
        for col in cols:
            self.__event_info =(col,item)
            self.event_generate('<<TreeviewInplaceEdit>>')
            if col in self._inplace_widgets:
                w = self._inplace_widgets[col]
                w.bind('<Key-Tab>',
                    lambda e: w.tk_focusNext().focus_set())
                w.bind('<Shift-Key-Tab>',
                    lambda e: w.tk_focusPrev().focus_set())

    def __updateWnds(self, event=None):
        if not self._curfocus:
            return
        item = self._curfocus
        cols = self.__get_display_columns()
        for col in cols:
            if col in self._inplace_widgets:
                wnd = self._inplace_widgets[col]
                bbox = ''
                if self.exists(item):
                    bbox = self.bbox(item, column=col)
                if bbox == '':
                    wnd.place_forget()
                elif col in self._inplace_widgets_show:
                    wnd.place(x=bbox[0], y=bbox[1],
                        width=bbox[2], height=bbox[3])

    def __clear_inplace_widgets(self):
        """Remove all inplace edit widgets."""
        cols = self.__get_display_columns()
        #print('Clear:', cols)
        for c in cols:
            if c in self._inplace_widgets:
                widget = self._inplace_widgets[c]
                widget.place_forget()
                self._inplace_widgets_show.pop(c, None)
                #widget.destroy()
                #del self._inplace_widgets[c]

    def __get_display_columns(self):
        cols = self.cget('displaycolumns')
        show = (str(s) for s in self.cget('show'))
        if '#all' in cols:
            cols = self.cget('columns') + ('#0',)
        elif 'tree' in show:
            cols = cols + ('#0',)
        return cols

    def get_event_info(self):
        return self.__event_info;

    def __get_value(self, col, item):
        if col == '#0':
            return self.item(item, 'text')
        else:
            return self.set(item, col)

    def __set_value(self, col, item, value):
        if col == '#0':
            self.item(item, text=value)
        else:
            self.set(item, col, value)
        self.__event_info =(col,item)
        self.event_generate('<<TreeviewCellEdited>>')

    def __update_value(self, col, item):
        if not self.exists(item):
            return
        value = self.__get_value(col, item)
        newvalue = self._inplace_vars[col].get()
        if value != newvalue:
            self.__set_value(col, item, newvalue)


    def inplace_entry(self, col, item):
        if col not in self._inplace_vars:
            self._inplace_vars[col] = tk.StringVar()
        svar = self._inplace_vars[col]
        svar.set(self.__get_value(col, item))
        if col not in self._inplace_widgets:
            self._inplace_widgets[col] = ttk.Entry(self, textvariable=svar)
        entry = self._inplace_widgets[col]
        entry.bind('<Unmap>', lambda e: self.__update_value(col, item))
        entry.bind('<FocusOut>', lambda e: self.__update_value(col, item))
        self._inplace_widgets_show[col] = True

    def inplace_checkbutton(self, col, item, onvalue='True', offvalue='False'):
        if col not in self._inplace_vars:
            self._inplace_vars[col] = tk.StringVar()
        svar = self._inplace_vars[col]
        svar.set(self.__get_value(col, item))
        if col not in self._inplace_widgets:
            self._inplace_widgets[col] = ttk.Checkbutton(self,
            textvariable=svar, variable=svar, onvalue=onvalue, offvalue=offvalue)
        cb = self._inplace_widgets[col]
        cb.bind('<Unmap>', lambda e: self.__update_value(col, item))
        cb.bind('<FocusOut>', lambda e: self.__update_value(col, item))
        self._inplace_widgets_show[col] = True

    def inplace_combobox(self, col, item, values, readonly=True,
                         update_values=False):
        state = 'readonly' if readonly else 'normal'
        if col not in self._inplace_vars:
            self._inplace_vars[col] = tk.StringVar()
        svar = self._inplace_vars[col]
        svar.set(self.__get_value(col, item))
        if col not in self._inplace_widgets:
            self._inplace_widgets[col] = ttk.Combobox(self,
                textvariable=svar, values=values, state=state)
        if update_values:
            self._inplace_widgets[col].configure(values=values)
        cb = self._inplace_widgets[col]
        cb.bind('<Unmap>', lambda e: self.__update_value(col, item))
        cb.bind('<FocusOut>', lambda e: self.__update_value(col, item))
        self._inplace_widgets_show[col] = True

    def inplace_spinbox(self, col, item, min, max, step):
        if col not in self._inplace_vars:
            self._inplace_vars[col] = tk.StringVar()
        svar = self._inplace_vars[col]
        svar.set(self.__get_value(col, item))
        if col not in self._inplace_widgets:
            self._inplace_widgets[col] = tk.Spinbox(self,
                textvariable=svar, from_=min, to=max, increment=step)
        sb = self._inplace_widgets[col]
        sb.bind('<Unmap>', lambda e: self.__update_value(col, item))
        sb.bind('<FocusOut>', lambda e: self.__update_value(col, item))
        self._inplace_widgets_show[col] = True


    def inplace_custom(self, col, item, widget, stringvar=None):
        if col not in self._inplace_vars:
            if stringvar is None:
                self._inplace_vars[col] = tk.StringVar()
            else:
                self._inplace_vars[col] = stringvar
        svar = self._inplace_vars[col]
        svar.set(self.__get_value(col, item))
        self._inplace_widgets[col] = widget
        widget.bind('<Unmap>', lambda e: self.__update_value(col, item))
        widget.bind('<FocusOut>', lambda e: self.__update_value(col, item))
        self._inplace_widgets_show[col] = True




# encoding: UTF-8
from __future__ import unicode_literals
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk


class Dialog(object):
    """
    Virtual events:
        <<DialogClose>>
    """

    def __init__(self, parent, modal=False):
        self.parent = parent
        self.is_modal = modal
        self.show_centered = True
        self.running_modal = False
        self.toplevel = tk.Toplevel(parent)
        self.toplevel.withdraw()
        self.toplevel.protocol("WM_DELETE_WINDOW", self._on_wm_delete_window)
        self.bind('<<DialogClose>>', self._default_close_action)

        self._init_before()
        self._create_ui()
        self._init_after()

        #self.frame.pack(fill='both', expand=True)


    def _init_before(self):
        pass

    def _create_ui(self):
        pass

    def _init_after(self):
        pass

    def set_modal(self, modal):
        self.is_modal = modal
    
    def center_window(self):
        """Center a window on its parent or screen."""
        window = self.toplevel
        height = window.winfo_height()
        width = window.winfo_width()
        parent = self.parent
        if parent:
            x_coord = int(parent.winfo_x() + (parent.winfo_width() / 2 - width / 2))
            y_coord = int(parent.winfo_y() + (parent.winfo_height() / 2 - height / 2))
        else:
            x_coord = int(window.winfo_screenwidth() / 2 - width / 2)
            y_coord = int(window.winfo_screenheight() / 2 - height / 2)
        geom = '{0}x{1}+{2}+{3}'.format(width, height, x_coord, y_coord)
        window.geometry(geom)
    
    def run(self):
        self.toplevel.transient(self.parent)
        self.toplevel.deiconify()
        self.toplevel.wait_visibility()
        if self.show_centered:
            self.center_window()
        initial_focus = self.toplevel.focus_lastfor()
        if initial_focus:
            initial_focus.focus_set()
        if self.is_modal:
            self.running_modal =True
            self.toplevel.grab_set()
    
    def destroy(self):
        if self.toplevel:
            self.toplevel.destroy()
        self.toplevel = None

    def _on_wm_delete_window(self):
        self.toplevel.event_generate('<<DialogClose>>')
    
    def _default_close_action(self, dialog):
        self.close()
    
    def close(self):
        if self.running_modal:
            self.toplevel.grab_release()
            self.running_modal = False
        self.toplevel.withdraw()
        self.parent.focus_set()
    
    def show(self):
        if self.toplevel:
            self.toplevel.deiconify()
    
    def set_title(self, title):
        """Sets the dialog title"""
        if self.toplevel:
            self.toplevel.title(title)

    #
    # interface to toplevel methods used by gui builder
    #
    def configure(self, cnf=None, **kw):
        if 'modal' in kw:
            value = kw.pop('modal')
            self.set_modal(tk.getboolean(value))
        self.toplevel.configure(cnf, **kw)
    
    config = configure

    def cget(self, key):
        if key == 'modal':
            return self.is_modal
        return self.toplevel.cget(key)

    __getitem__ = cget

    def __setitem__(self, key, value):
        self.configure({key: value})

    def bind(self, sequence=None, func=None, add=None):
        def dialog_cb(event, dialog=self):
            func(dialog)
        return self.toplevel.bind(sequence, dialog_cb, add)
        
#    def unbind(self, sequence, funcid=None):
#        pass


if __name__ == '__main__':
    class TestDialog(Dialog):
        def _create_ui(self):
            label = ttk.Label(self.frame, text='TestDialog Class')
            label.pack()
            entry = ttk.Entry(self.frame)
            entry.pack()
#            f = ttk.Frame(self.toplevel)
#            self.default_buttonbox(f)
#            f.pack(fill='x')

    app = tk.Tk()
    dialog = None

    def show_dialog():
        global dialog
        if dialog is None:
            dialog = TestDialog(app)
            dialog.run()
        else:
            dialog.show()
            
    def custom_callback(dialog):
        print('Custom callback')
        dialog.close()

    def show_modal_dialog():
        dialog = TestDialog(app)
        dialog.set_title('Modal dialog')
        dialog.set_modal(True)
        dialog.bind('<<DialogClose>>', custom_callback)
        print('before run')
        dialog.run()
        print('after run')

    btn = tk.Button(app, text='show dialog', command=show_dialog)
    btn.pack()
    btn = tk.Button(app, text='show modal dialog', command=show_modal_dialog)
    btn.pack()

    app.mainloop()



# encoding: utf8
from __future__ import unicode_literals

try:
    import tkinter as tk
    from tkinter.scrolledtext import ScrolledText
except:
    import Tkinter as tk
    from ScrolledText import ScrolledText


from pygubu.builder.builderobject import BuilderObject, register_widget
from pygubu.builder.tkstdwidgets import TKText


class TkinterScrolledTextBO(TKText):
    class_ = ScrolledText


register_widget('pygubu.builder.widgets.tkinterscrolledtext',
                TkinterScrolledTextBO,
                'ScrolledText', ('Control & Display', 'tk'))



# encoding: utf8
from __future__ import unicode_literals

__all__ = ['Builder', 'TkApplication', 'BuilderObject', 'register_widget',
           'register_property', 'remove_binding', 'ApplicationLevelBindManager']

import pygubu.builder.builderobject
from pygubu.builder import Builder
from pygubu.builder.builderobject import BuilderObject, register_widget
from pygubu.binding import remove_binding, ApplicationLevelBindManager


__version__ = '0.10.8'


def register_property(name, description):
    return pygubu.builder.builderobject.register_property(name, description)



class TkApplication:
    def __init__(self, master=None):
        self.master = master
        self.toplevel = master.winfo_toplevel()

        self.toplevel.withdraw()
        self._init_before()
        self._create_ui()
        self._init_after()
        self.toplevel.deiconify()

    def _init_before(self):
        pass

    def _create_ui(self):
        pass

    def _init_after(self):
        pass

    def run(self):
        """Ejecute the main loop."""

        self.toplevel.protocol("WM_DELETE_WINDOW", self.__on_window_close)
        self.toplevel.mainloop()

    def set_resizable(self):
        self.toplevel.rowconfigure(0, weight=1)
        self.toplevel.columnconfigure(0, weight=1)

    def set_title(self, title):
        """Set the window title."""
        self.toplevel.title(title)

    def set_menu(self, menu):
        """Set the main menu."""
        self.toplevel.config(menu=menu)

    def __on_window_close(self):
        """Manage WM_DELETE_WINDOW protocol."""
        if self.on_close_execute():
            self.toplevel.destroy()

    def on_close_execute(self):
        """Determine if if the application is ready for quit,
           return boolean."""
        return True

    def quit(self):
        """Exit the app if it is ready for quit."""
        self.__on_window_close()

    def set_size(self, geom):
        self.toplevel.geometry(geom)



# encoding: utf8
from pygubu import BuilderObject, register_widget, register_property
from pygubu.widgets.pathchooserinput import PathChooserInput


class PathChooserInputBuilder(BuilderObject):
    class_ = PathChooserInput
    OPTIONS_CUSTOM = ('type', 'path', 'image', 'textvariable', 'state',
                      'initialdir', 'mustexist', 'title',)
    properties = OPTIONS_CUSTOM
    virtual_events = ('<<PathChooserPathChanged>>',)
    
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == 'type':
            code_bag[pname] = "'{0}'".format(value)
        elif pname in ('initialdir', 'mustexist', 'title'):
            code_bag[pname] = "'{0}'".format(value)
        else:
            super(PathChooserInputBuilder, self)._code_set_property(
                targetid, pname, value, code_bag)


register_widget('pygubu.builder.widgets.pathchooserinput', PathChooserInputBuilder,
                'PathChooserInput', ('ttk', 'Pygubu Widgets'))

props = {
    'type': {
        'editor': 'choice',
        'params': {
            'values': (PathChooserInput.FILE, PathChooserInput.DIR), 'state': 'readonly'},
        'default': PathChooserInput.FILE
        },
    'path': {
        'editor': 'entry'
        },
    'image': {
        'editor': 'imageentry'
        },
    'textvariable': {
        'editor': 'tkvarentry'
        },
    'state': {
        'editor': 'choice',
        'pygubu.builder.widgets.pathchooserinput': {
            'params': {
                'values': ('', 'normal', 'disabled', 'readonly'),
                'state': 'readonly'}},
        },
    'mustexist': {
        'editor': 'dynamic',
        'params': {
            'mode': 'choice',
            'values': ('', 'false', 'true'), 'state': 'readonly'},
        },
    'initialdir': {
        'editor': 'dynamic',
        'params': {'mode': 'entry'},
        },
    }

for p in props:
    register_property(p, props[p])




# encoding: utf8
from __future__ import unicode_literals

try:
    import tkinter as tk
except:
    import Tkinter as tk

from pygubu import ApplicationLevelBindManager as BindManager
from pygubu.binding import remove_binding


class TkScrolledFrame(tk.Frame):
    VERTICAL = 'vertical'
    HORIZONTAL = 'horizontal'
    BOTH = 'both'
    _framecls = tk.Frame
    _sbarcls = tk.Scrollbar

    def __init__(self, master=None, **kw):
        self.scrolltype = kw.pop('scrolltype', self.VERTICAL)
        self.usemousewheel = tk.getboolean(kw.pop('usemousewheel', False))
        self._bindingids = []

        #super(ScrolledFrameBase, self).__init__(master, **kw)
        self._framecls.__init__(self, master, **kw)

        self._container = self._framecls(self, width=200, height=200)
        self._clipper = self._framecls(self._container, width=200, height=200)
        self.innerframe = innerframe = self._framecls(self._clipper)
        self.vsb = vsb = self._sbarcls(self._container)
        self.hsb = hsb = self._sbarcls(self._container, orient="horizontal")

        # variables
        self.hsbOn = 0
        self.vsbOn = 0
        self.hsbNeeded = 0
        self.vsbNeeded = 0
        self._jfraction=0.05
        self._scrollTimer = None
        self._scrollRecurse = 0
        self._startX = 0
        self._startY = 0

        #configure scroll
        self.hsb.set(0.0, 1.0)
        self.vsb.set(0.0, 1.0)
        self.vsb.config(command=self.yview)
        self.hsb.config(command=self.xview)

        #grid
        self._container.pack(expand=True, fill='both')
        self._clipper.grid(row=0, column=0, sticky=tk.NSEW)
        #self.vsb.grid(row=0, column=1, sticky=tk.NS)
        #self.hsb.grid(row=1, column=0, sticky=tk.EW)
        
        self._container.rowconfigure(0, weight=1)
        self._container.columnconfigure(0, weight=1)

        # Whenever the clipping window or scrolled frame change size,
        # update the scrollbars.
        self.innerframe.bind('<Configure>', self._reposition)
        self._clipper.bind('<Configure>', self._reposition)
        self.bind('<Configure>', self._reposition)
        self._configure_mousewheel()

    # Set timer to call real reposition method, so that it is not
    # called multiple times when many things are reconfigured at the
    # same time.
    def reposition(self):
        if self._scrollTimer is None:
            self._scrollTimer = self.after_idle(self._scrollBothNow)

    # Called when the user clicks in the horizontal scrollbar.
    # Calculates new position of frame then calls reposition() to
    # update the frame and the scrollbar.
    def xview(self, mode = None, value = None, units = None):
        if type(value) == str:
            value = float(value)
        if mode is None:
            return self.hsb.get()
        elif mode == 'moveto':
            frameWidth = self.innerframe.winfo_reqwidth()
            self._startX = value * float(frameWidth)
        else: # mode == 'scroll'
            clipperWidth = self._clipper.winfo_width()
            if units == 'units':
                jump = int(clipperWidth * self._jfraction)
            else:
                jump = clipperWidth
            self._startX = self._startX + value * jump

        self.reposition()

    # Called when the user clicks in the vertical scrollbar.
    # Calculates new position of frame then calls reposition() to
    # update the frame and the scrollbar.
    def yview(self, mode = None, value = None, units = None):

        if type(value) == str:
            value = float(value)
        if mode is None:
            return self.vsb.get()
        elif mode == 'moveto':
            frameHeight = self.innerframe.winfo_reqheight()
            self._startY = value * float(frameHeight)
        else: # mode == 'scroll'
            clipperHeight = self._clipper.winfo_height()
            if units == 'units':
                jump = int(clipperHeight * self._jfraction)
            else:
                jump = clipperHeight
            self._startY = self._startY + value * jump

        self.reposition()

    def _reposition(self, event):
        self.reposition()

    def _getxview(self):

        # Horizontal dimension.
        clipperWidth = self._clipper.winfo_width()
        frameWidth = self.innerframe.winfo_reqwidth()
        if frameWidth <= clipperWidth:
            # The scrolled frame is smaller than the clipping window.

            self._startX = 0
            endScrollX = 1.0
            #use expand by default
            relwidth = 1
        else:
            # The scrolled frame is larger than the clipping window.
            #use expand by default
            if self._startX + clipperWidth > frameWidth:
                self._startX = frameWidth - clipperWidth
                endScrollX = 1.0
            else:
                if self._startX < 0:
                    self._startX = 0
                endScrollX = (self._startX + clipperWidth) / float(frameWidth)
            relwidth = ''

        # Position frame relative to clipper.
        self.innerframe.place(x = -self._startX, relwidth = relwidth)
        return (self._startX / float(frameWidth), endScrollX)

    def _getyview(self):

        # Vertical dimension.
        clipperHeight = self._clipper.winfo_height()
        frameHeight = self.innerframe.winfo_reqheight()
        if frameHeight <= clipperHeight:
            # The scrolled frame is smaller than the clipping window.

            self._startY = 0
            endScrollY = 1.0
            # use expand by default
            relheight = 1
        else:
            # The scrolled frame is larger than the clipping window.
            # use expand by default
            if self._startY + clipperHeight > frameHeight:
                self._startY = frameHeight - clipperHeight
                endScrollY = 1.0
            else:
                if self._startY < 0:
                    self._startY = 0
                endScrollY = (self._startY + clipperHeight) / float(frameHeight)
            relheight = ''

        # Position frame relative to clipper.
        self.innerframe.place(y = -self._startY, relheight = relheight)
        return (self._startY / float(frameHeight), endScrollY)

    # According to the relative geometries of the frame and the
    # clipper, reposition the frame within the clipper and reset the
    # scrollbars.
    def _scrollBothNow(self):
        self._scrollTimer = None

        # Call update_idletasks to make sure that the containing frame
        # has been resized before we attempt to set the scrollbars.
        # Otherwise the scrollbars may be mapped/unmapped continuously.
        self._scrollRecurse = self._scrollRecurse + 1
        self.update_idletasks()
        self._scrollRecurse = self._scrollRecurse - 1
        if self._scrollRecurse != 0:
            return

        xview = self._getxview()
        yview = self._getyview()
        self.hsb.set(xview[0], xview[1])
        self.vsb.set(yview[0], yview[1])

        require_hsb = self.scrolltype in (self.BOTH, self.HORIZONTAL)
        self.hsbNeeded = (xview != (0.0, 1.0)) and require_hsb
        require_vsb = self.scrolltype in (self.BOTH, self.VERTICAL)
        self.vsbNeeded = (yview != (0.0, 1.0)) and require_vsb

        # If both horizontal and vertical scrollmodes are dynamic and
        # currently only one scrollbar is mapped and both should be
        # toggled, then unmap the mapped scrollbar.  This prevents a
        # continuous mapping and unmapping of the scrollbars.
        if (self.hsbNeeded != self.hsbOn and
            self.vsbNeeded != self.vsbOn and
            self.vsbOn != self.hsbOn):
            if self.hsbOn:
                self._toggleHorizScrollbar()
            else:
                self._toggleVertScrollbar()
            return

        if self.hsbNeeded != self.hsbOn:
            self._toggleHorizScrollbar()

        if self.vsbNeeded != self.vsbOn:
            self._toggleVertScrollbar()

    def _toggleHorizScrollbar(self):

        self.hsbOn = not self.hsbOn

        #interior = self #.origInterior
        if self.hsbOn:
            self.hsb.grid(row=1, column=0, sticky=tk.EW)
            #interior.grid_rowconfigure(3, minsize = self['scrollmargin'])
        else:
            self.hsb.grid_forget()
            #interior.grid_rowconfigure(3, minsize = 0)

    def _toggleVertScrollbar(self):

        self.vsbOn = not self.vsbOn

        #interior = self#.origInterior
        if self.vsbOn:
            self.vsb.grid(row=0, column=1, sticky=tk.NS)
            #interior.grid_columnconfigure(3, minsize = self['scrollmargin'])
        else:
            self.vsb.grid_forget()
            #interior.grid_columnconfigure(3, minsize = 0)

    def configure(self, cnf=None, **kw):
        args = tk._cnfmerge((cnf, kw))
        key = 'usemousewheel'
        if key in args:
            self.usemousewheel = tk.getboolean(args[key])
            del args[key]
            self._configure_mousewheel()
        #super(ScrolledFrameBase, self).configure(args)
        self._framecls.configure(self, args)

    config = configure

    def cget(self, key):
        option = 'usemousewheel'
        if key == option:
            return self.usemousewheel
        #return super(ScrolledFrameBase, self).cget(key)
        return self._framecls.cget(self, key)

    __getitem__ = cget

    def _configure_mousewheel(self):
        if self.usemousewheel:
            BindManager.init_mousewheel_binding(self)

            if self.hsb and not hasattr(self.hsb, 'on_mousewheel'):
                self.hsb.on_mousewheel = BindManager.make_onmousewheel_cb(self, 'x', 2)
            if self.vsb and not hasattr(self.vsb, 'on_mousewheel'):
                self.vsb.on_mousewheel = BindManager.make_onmousewheel_cb(self, 'y', 2)

            main_sb = self.vsb or self.hsb
            if main_sb:
                self.on_mousewheel = main_sb.on_mousewheel
                bid = self.bind('<Enter>',
                                lambda event: BindManager.mousewheel_bind(self),
                                add='+')
                self._bindingids.append((self, bid))
                bid = self.bind('<Leave>',
                                lambda event: BindManager.mousewheel_unbind(),
                                add='+')
                self._bindingids.append((self, bid))
            for s in (self.vsb, self.hsb):
                if s:
                    bid = s.bind('<Enter>',
                                 lambda event, scrollbar=s: BindManager.mousewheel_bind(scrollbar),
                                 add='+')
                    self._bindingids.append((s, bid))
                    if s != main_sb:
                        bid = s.bind('<Leave>',
                                     lambda event: BindManager.mousewheel_unbind(),
                                     add='+')
                        self._bindingids.append((s, bid))
        else:
            for widget, bid in self._bindingids:
                remove_binding(widget, bid)


#class TkScrolledFrameFactory(type):
#    def __new__(cls, clsname, superclasses, attrs):
#        return type.__new__(cls, str(clsname), superclasses, attrs)


#TkScrolledFrame = TkScrolledFrameFactory('TkScrolledFrame',
#                                       (ScrolledFrameBase, tk.Frame, object),
#                                       {'_framecls':tk.Frame,
#                                        '_sbarcls': tk.Scrollbar})



# encoding: UTF-8
from __future__ import unicode_literals, print_function
import xml.etree.ElementTree as ET
from collections import namedtuple

__all__ = ['WidgetMeta', 'BindingMeta']

BindingMeta = namedtuple('BindingMeta',
                         ['sequence', 'handler', 'add'])

GridRCLine = namedtuple('GridRCLine', ['rctype', 'rcid', 'pname', 'pvalue'])


class WidgetMeta(object):
    def __init__(self, cname, identifier, manager=None,
                 properties_defaults=None,
                 layout_defaults=None):
        super(WidgetMeta, self).__init__()
        self.classname = cname
        self.identifier = identifier
        self.properties = {}
        self.bindings = []
        self._manager = manager if manager is not None else 'grid'
        self.layout_required = True
        self.layout_properties = {}
        self.gridrc_properties = []
        self.properties_defaults = properties_defaults if properties_defaults \
            is not None else {}
        self.layout_defaults = layout_defaults if layout_defaults \
            is not None else {}
        
        # init defaults
        self.apply_properties_defaults()
        self.apply_layout_defaults()
    
    @property
    def manager(self):
        return self._manager
    
    @manager.setter
    def manager(self, value):
        self._manager = value
        
    def apply_properties_defaults(self):
        for name, value in self.properties_defaults.items():
            self.properties[name] = value
    
    def apply_layout_defaults(self):
        layout_defaults_by_manager = self.layout_defaults
        if self.manager in self.layout_defaults:
            layout_defaults_by_manager = self.layout_defaults[self.manager]
        for name, value in layout_defaults_by_manager.items():
            self.layout_properties[name] = value
    
    def has_layout_defined(self):
        return (len(self.layout_properties) > 0)
    
    def clear_layout(self):
        self.layout_properties = {}
        self.gridrc_properties = []
        self.apply_layout_defaults()
    
    def get_gridrc_value(self, rctype, rcid, pname):
        value = None
        for line in self.gridrc_properties:
            if (line.rctype == rctype and line.rcid == rcid
                and line.pname == pname):
                value = line.pvalue
                break
        return value
    
    def set_gridrc_value(self, rctype, rcid, pname, value):
        index = None
        for i, r in enumerate(self.gridrc_properties):
            if r.rctype == rctype and r.rcid == rcid and r.pname == pname:
                index = i
                break
        if index is None:
            line = GridRCLine(rctype, rcid, pname, value)
            self.gridrc_properties.append(line)
        else:
            line = GridRCLine(rctype, rcid, pname, value)
            self.gridrc_properties[index] = line
    
    def __repr__(self):
        tpl = '''<WidgetMeta classname: {0} identifier: {1}>'''
        return tpl.format(self.classname, self.identifier)
    
    def copy_gridrc(self, from_, rctype):
        '''Copy gridrc lines of type rctype from from_'''
        rc = [ line for line in self.gridrc_properties 
               if line.rctype != rctype ]
        for line in from_.gridrc_properties:
            if line.rctype == rctype:
                rc.append(line)
        self.gridrc_properties = rc
    
    def copy_properties(self, wfrom):
        self.properties = wfrom.properties.copy()


if __name__ == '__main__':
    w = WidgetMeta('mywidget', 'w1')
    w.manager = 'pack'
    w.properties = {
        'key1': 'value1',
        'key2': 'value2'
    }
    w.bindings = [
        BindingMeta('<<event_1>>', 'callback_1', ''),
        BindingMeta('<<event_2>>', 'callback_2', '+')
    ]
    w.layout_properties = {
        'prop1': 'value1',
        'prop2': 'value2'
    }
    
    print('To xml')
    node = w.to_xmlnode()
    node = ET.tostring(node)
    print(node, end='\n\n')
    
    print('From xml:')
    strdata = '''<?xml version='1.0' encoding='utf-8'?>
<object class="ttk.Label" id="Label_1">
        <property name="text" translatable="yes">Label_1</property>
        <property name="background">yellow</property>
        <layout manager="grid">
          <property name="propagate">True</property>
          <property type="row" id="0" name="pad">20</property>
          <property type="col" id="0" name="pad">10</property>
          <property type="row" id="0" name="weight">1</property>
        </layout>
      </object>
    '''
    node = ET.fromstring(strdata)
    meta = WidgetMeta.from_xmlnode(node)
    print(meta)


# encoding: utf8
try:
    import ttk
except:
    import tkinter.ttk as ttk

class Spinbox(ttk.Entry):

    def __init__(self, master=None, **kw):
        ttk.Entry.__init__(self, master, "ttk::spinbox", **kw)

    def current(self, newindex=None):
        return self.tk.call(self._w, 'current', newindex)

    def set(self, value):
        return self.tk.call(self._w, 'set', value)



# encoding: utf8
from __future__ import unicode_literals
try:
    import tkinter as tk
except:
    import Tkinter as tk

from pygubu.builder.builderobject import *
from pygubu.builder.tkstdwidgets import TKToplevel
from pygubu.widgets.dialog import Dialog


class DialogBO(TKToplevel):
    class_ = Dialog
    OPTIONS_STANDARD = TKToplevel.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TKToplevel.OPTIONS_SPECIFIC
    OPTIONS_CUSTOM = \
        TKToplevel.OPTIONS_CUSTOM + ('modal',)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    
    def layout(self, target=None):
        super(DialogBO, self).layout(self.widget.toplevel)

    def _set_property(self, target_widget, pname, value):
        if pname == 'modal':
            modal = False
            value = value.lower()
            if value == 'true':
                modal = True
            self.widget.set_modal(modal)
        else:
            super(DialogBO, self)._set_property(
                self.widget.toplevel, pname, value)

    def get_child_master(self):
        return self.widget.toplevel
    
    #
    # Code generation methods
    #
    def code_child_master(self):
        return '{0}.toplevel'.format(self.code_identifier())
    
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == 'modal':
            code_bag[pname] = "'{0}'".format(value)
        else:
            super(DialogBO, self)._code_set_property(targetid, pname,
                                                     value, code_bag)


register_widget('pygubu.builder.widgets.dialog', DialogBO,
                'Dialog', ('Pygubu Widgets', 'ttk'))


modal_prop = {
    'editor': 'choice',
    'params': {
        'values': ('true', 'false'), 'state': 'readonly'},
    'default': 'false'}

register_property('modal', modal_prop)



# encoding: utf8
from __future__ import unicode_literals

import os
import importlib
import logging
import xml.etree.ElementTree as ET
try:
    import tkinter
except:
    import Tkinter as tkinter

from pygubu.builder.builderobject import BuilderObject, CLASS_MAP, CB_TYPES
from pygubu.builder.widgetmeta import WidgetMeta
from pygubu.stockimage import StockImage, StockImageException
from .uidefinition import UIDefinition

logger = logging.getLogger(__name__)


#
# Builder class
#

class Builder(object):
    """Allows to build a tk interface from xml definition."""
    TK_VARIABLE_TYPES = ('string', 'int', 'boolean', 'double')

    def __init__(self, translator=None):
        super(Builder, self).__init__()
        self.uidefinition = UIDefinition(translator=translator)
        self.root = None
        self.objects = {}
        self.tkvariables = {}
        self._resource_paths = []
        self.translator = translator

    def add_resource_path(self, path):
        """Add additional path to the resources paths."""
        self._resource_paths.append(path)

    def get_image(self, path):
        """Return tk image corresponding to name which is taken form path."""
        image = ''
        name = os.path.basename(path)
        self.__load_image(path)
        try:
            image = StockImage.get(name)
        except StockImageException:
            # TODO: notify something here.
            pass
        return image
    
    def get_iconbitmap(self, path):
        """Return path to use as iconbitmap property."""
        image = None
        name = os.path.basename(path)
        self.__load_image(path)
        try:
            image = StockImage.as_iconbitmap(name)
        except StockImageException:
            # TODO: notify something here.
            pass
        return image
    
    def __load_image(self, path):
        name = os.path.basename(path)
        if not StockImage.is_registered(name):
            ipath = self.__find_image(path)
            if ipath is not None:
                StockImage.register(name, ipath)
            else:
                msg = "Image '%s' not found in resource paths."
                logger.warning(msg, name)        

    def __find_image(self, relpath):
        image_path = None
        for rp in self._resource_paths:
            for root, dirs, files in os.walk(rp):
                if relpath in files:
                    image_path = os.path.join(root, relpath)
                    break
            if image_path is not None:
                break
        return image_path

    def get_variable(self, varname):
        """Return a tk variable created with 'create_variable' method."""
        return self.tkvariables[varname]
        
    def import_variables(self, container, varnames=None):
        """Helper method to avoid call get_variable for every variable."""
        if varnames is None:
            for keyword in self.tkvariables:
                setattr(container, keyword, self.tkvariables[keyword])
        else:
            for keyword in varnames:
                if keyword in self.tkvariables:
                    setattr(container, keyword, self.tkvariables[keyword])

    def _process_variable_description(self, name_or_desc):
        vname = name_or_desc
        vtype = 'string'  # default type if not defined
        if ':' in name_or_desc:
            vtype, vname = name_or_desc.split(':')
            #  Fix incorrect order bug #33
            if vtype not in self.TK_VARIABLE_TYPES:
                #  Swap order
                vtype, vname = vname, vtype
                if vtype not in self.TK_VARIABLE_TYPES:
                    msg = 'Undefined variable type in "{0}"'.format(vname)
                    raise Exception(msg)
        return (vname, vtype)

    def create_variable(self, varname, vtype=None):
        """Create a tk variable.
        If the variable was created previously return that instance.
        """
        vname, type_from_name = self._process_variable_description(varname)

        if vname in self.tkvariables:
            var = self.tkvariables[vname]
        else:
            if vtype is None:
                # get type from name
                if type_from_name == 'int':
                    var = tkinter.IntVar()
                elif type_from_name == 'boolean':
                    var = tkinter.BooleanVar()
                elif type_from_name == 'double':
                    var = tkinter.DoubleVar()
                else:
                    var = tkinter.StringVar()
            else:
                var = vtype()

            self.tkvariables[vname] = var
        return var

    def add_from_file(self, file_or_filename):
        """Load ui definition from file."""
        self.uidefinition.load_file(file_or_filename)

    def add_from_string(self, strdata):
        self.uidefinition.load_from_string(strdata)

    def add_from_xmlnode(self, element):
        """Load ui definition from xml.etree.Element node."""
        self.uidefinition.add_xmlnode(element)

    def get_object(self, name, master=None):
        """Find and create the widget named name.
        Use master as parent. If widget was already created, return
        that instance."""
        widget = None
        if name in self.objects:
            widget = self.objects[name].widget
        else:
            wmeta = self.uidefinition.get_widget(name)
            if wmeta is not None:
                rmeta = WidgetMeta('root', 'root')
                root = BuilderObject(self, rmeta)
                root.widget = master
                bobject = self._realize(root, wmeta)
                widget = bobject.widget
        if widget is None:
            msg = 'Widget "{0}" not defined.'.format(name)
            raise Exception(msg)
        return widget

    def _import_class(self, modulename):
        if modulename.startswith('tk.'):
            importlib.import_module('pygubu.builder.tkstdwidgets')
        elif modulename.startswith('ttk.'):
            importlib.import_module('pygubu.builder.ttkstdwidgets')
        else:
            # Import module as full path
            try:
                importlib.import_module(modulename)
            except ImportError as e:
                msg = 'Failed to import module as fullname: %s'
                logger.warning(msg, modulename)
                logger.exception(e)
                # A single module can contain various widgets
                # try to import the first part of the path
                if '.' in modulename:
                    first, last = modulename.rsplit('.', 1)
                    try:
                        importlib.import_module(first)
                    except ImportError as e:
                        importlib.import_module(last)
                else:
                    raise e

    def _realize(self, master, wmeta):
        """Builds a widget from widget metadata using master as parent."""
        
        if wmeta.classname not in CLASS_MAP:
            self._import_class(wmeta.classname)

        if wmeta.classname in CLASS_MAP:
            bclass = CLASS_MAP[wmeta.classname].builder
            parent = bclass.factory(self, wmeta)
            self._pre_realize(parent)
            parent.realize(master)

            self.objects[wmeta.identifier] = parent

            for childmeta in \
                self.uidefinition.widget_children(wmeta.identifier):
                child = self._realize(parent, childmeta)
                parent.add_child(child)
            parent.configure()
            parent.layout()
            
            self._post_realize(parent)
            
            return parent
        else:
            msg = 'Class "{0}" not mapped'.format(wmeta.classname)
            raise Exception(msg)

    def _pre_realize(self, bobject):
        wmeta = bobject.wmeta
        cname = wmeta.classname
        wmeta.layout_required = bobject.layout_required
        has_layout = (len(wmeta.layout_properties) > 1)
        if wmeta.layout_required and not has_layout:
            logger.debug('No layout information for: (%s, %s).',
                           cname, wmeta.identifier)
    
    def _post_realize(self, bobject):
        pass

    def connect_callbacks(self, callbacks_bag):
        """Connect callbacks specified in callbacks_bag with callbacks
        defined in the ui definition.
        Return a list with the name of the callbacks not connected.
        """
        notconnected = []
        for wname, builderobj in self.objects.items():
            missing = builderobj.connect_commands(callbacks_bag)
            if missing is not None:
                notconnected.extend(missing)
            missing = builderobj.connect_bindings(callbacks_bag)
            if missing is not None:
                notconnected.extend(missing)
        if notconnected:
            notconnected = list(set(notconnected))
            msg = 'Missing callbacks for commands: %s'
            logger.warning(msg, notconnected)
            return notconnected
        else:
            return None
    
    def code_create_variable(self, name_or_desc, value, vtype=None):
        raise NotImplementedError()
    
    def code_create_image(self, filename):
        raise NotImplementedError()
    
    def code_create_iconbitmap(self, filename):
        raise NotImplementedError()
    
    def code_classname_for(self, bobject):
        raise NotImplementedError()
    
    def code_create_callback(self, widgetid, cbname, cbtype, args=None):
        raise NotImplementedError()



# encoding: UTF-8
from __future__ import unicode_literals, print_function
import sys
import json
import operator
import xml.etree.ElementTree as ET
from pygubu.builder.builderobject import CLASS_MAP
from pygubu.builder.widgetmeta import WidgetMeta, BindingMeta, GridRCLine


# in-place prettyprint formatter
def indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

try:
    JSONDecodeError = json.decoder.JSONDecodeError
except AttributeError:  # Python 2
    JSONDecodeError = ValueError


class UIDefinition(object):
    TRANSLATABLE_PROPERTIES = ['label', 'text', 'title']
    TK_COMMAND_PROPERTIES =(
        'command', 'validatecommand', 'invalidcommand', 'postcommand',
        'xscrollcommand', 'yscrollcommand', 'tearoffcommand',
    )
    
    def __init__(self, wmetaclass=None, translator=None):
        super(UIDefinition, self).__init__()
        self.tree = None
        self.root = None
        self._latest_version = '1.1'
        self.version = self._latest_version
        self.author = ''
        self._ignore_properties = (
            'command_id_arg', 'idtocommand',
            'validatecommand_args', 'invalidcommand_args',
        )
        self.wmetaclass = wmetaclass
        if wmetaclass is None:
            self.wmetaclass = WidgetMeta
        self.translator = translator
        self.__create()
    
    def _prop_from_xml(self, pnode, element):
        pname = pnode.get('name')
        pvalue = pnode.text
        
        if pname in self._ignore_properties:
            return (None, None)
        
        if self.translator is not None and pnode.get('translatable'):
            pvalue = self.translator(pvalue)
        
        # if node has a type property, send value as a json string
        jvalue = {}
        if pnode.get('type'):
            for attr, attrval in pnode.items():
                jvalue[attr] = attrval
            jvalue['value'] = pnode.text
            pvalue = json.dumps(jvalue)
        # Process old ui versions
        if self.version < '1.1':
            if pname in self.TK_COMMAND_PROPERTIES:
                cmd = {
                    'type': 'command',
                    'value': pnode.text,
                    'cbtype': 'simple',
                    'args': '',
                }
                # get old format value
                xpath = "./property[@name='{0}']"
                for oldp in ('command_id_arg', 'idtocommand'):
                    dpath = xpath.format(oldp)
                    node = element.find(dpath)
                    if node is not None:
                        nvalue = node.text.lower()
                        if nvalue == 'true':
                            cmd['cbtype'] = 'with_wid'
                
                if pname in ('validatecommand', 'invalidcommand'):
                    cmd['cbtype'] = 'entry_validate'
                    # get old format args
                    pargs = '{0}_args'.format(pname)
                    dpath = xpath.format(pargs)
                    node = element.find(dpath)
                    if node is not None:
                        cmd['args'] = node.text
                pvalue = json.dumps(cmd)
        return (pname, pvalue)
    
    def xmlnode_to_widget(self, element):
        elemid = element.get('id')
        meta = self.wmetaclass(element.get('class'), elemid)
    
        # properties
        properties = element.findall('./property')
        pdict = {}
        for p in properties:
            pname, pvalue = self._prop_from_xml(p, element)
            if pname is not None:
                pdict[pname] = pvalue
        
        meta.properties = pdict
    
        # Bindings
        bindings = []
        bind_elements = element.findall('./bind')
        for e in bind_elements:
            binding = BindingMeta(
                e.get('sequence'), e.get('handler'), e.get('add')
            )
            bindings.append(binding)
        meta.bindings = bindings
    
        # layout properties
        # use grid layout by default
        manager = 'grid'
        layout_elem = element.find('./layout')
        if layout_elem is not None:
            manager = layout_elem.get('manager', 'grid')
            meta.manager = manager
            props = layout_elem.findall('./property')
            if manager == 'grid':
                for p in props:
                    ptype = p.get('type', None)
                    if ptype is None:
                        meta.layout_properties[p.get('name')] = p.text
                    else:
                        rcid = p.get('id')
                        rcname = p.get('name')
                        rcvalue = p.text
                        line = GridRCLine(ptype, rcid, rcname, rcvalue)
                        meta.gridrc_properties.append(line)
            else:
                for p in props:
                    meta.layout_properties[p.get('name')] = p.text
        if self.version == '':
            # try to load old version grid rc info
            # Gridrc info was in the parent, or
            # in the widget itself if has no parent.
            xpath = ".//*[@id='{0}']/../..".format(elemid)
            parent = self.root.find(xpath)
            if parent is None:
                if layout_elem is not None:
                    self.__load_old_gridrc_layout(layout_elem, meta)
            else:
                layout_elem = parent.find('./layout')
                if layout_elem is not None:
                    self.__load_old_gridrc_layout(layout_elem, meta)
        
        return meta
    
    def __load_old_gridrc_layout(self, element, meta):
        '''Load old grid rc information.'''
        
        rows = element.findall('./rows/row')
        for row in rows:
            rid = row.get('id')
            props = row.findall('./property')
            for p in props:
                rpname = p.get('name')
                rpvalue = p.text
                line = GridRCLine('row', rid, rpname, rpvalue)
                meta.gridrc_properties.append(line)
        columns = element.findall('./columns/column')
        for col in columns:
            cid = col.get('id')
            props = col.findall('./property')
            for p in props:
                cpname = p.get('name')
                cpvalue = p.text
                line = GridRCLine('col', cid, cpname, cpvalue)
                meta.gridrc_properties.append(line)
        
    def _prop_to_xml(self, pname, pvalue):
        pnode = ET.Element('property')
        pnode.set('name', pname)
        pnode.text = pvalue
        if pname in self.TRANSLATABLE_PROPERTIES:
            pnode.set('translatable', 'yes')
        # if pvalue is a json do special
        try:
            dv = json.loads(pvalue)
            if isinstance(dv, dict):
                if 'value' not in dv:
                    raise Exception('Invalid json value for property')
                for k, attrval in dv.items():
                    if k != 'value':
                        pnode.set(k, str(attrval))
                pnode.text = dv['value']
        except JSONDecodeError:
            pass
        
        return pnode
    
    def widget_to_xmlnode(self, wmeta):
        '''Returns xml representation of widget'''
        
        node = ET.Element('object')
    
        node.set('class', wmeta.classname)
        node.set('id', wmeta.identifier)
    
        pkeys = sorted(wmeta.properties.keys())
        for pkey in pkeys:
            pnode = self._prop_to_xml(pkey, wmeta.properties[pkey])
            node.append(pnode)
    
        # bindings:
        bindings = sorted(wmeta.bindings, key=operator.itemgetter(0, 1))
        for b in bindings:
            bind = ET.Element('bind')
            for key in b._fields:
                bind.set(key, getattr(b, key))
            node.append(bind)
    
        # layout:
        #if self.layout_required:
        if CLASS_MAP[wmeta.classname].builder.layout_required:
            # create layout node
            layout_node = ET.Element('layout')
            layout_node.set('manager', wmeta.manager)
            
            keys = sorted(wmeta.layout_properties)
            for prop in keys:
                pnode = ET.Element('property')
                pnode.set('name', prop)
                pnode.text = wmeta.layout_properties[prop]
                layout_node.append(pnode)
            
            lines = sorted(wmeta.gridrc_properties,
                           key=operator.itemgetter(0,1,2))
            for line in lines:
                p = ET.Element('property')
                p.set('type', line.rctype)
                p.set('id', line.rcid)
                p.set('name', line.pname)
                p.text = line.pvalue
                layout_node.append(p)
            # Append node layout
            node.append(layout_node)
        return node    
        
    def __create(self):
        # Version 1.0: start of schema versioning, implements multiple layout managers
        # Version 1.1: remove idtocommand and command_id_arg properties
        self.root = root = ET.Element('interface')
        root.set('version', self._latest_version)
        if self.author:
            root.set('author', self.author)
        self.tree = ET.ElementTree(root)
    
    def _tree_load(self, tree, default_version=None):
        if default_version is None:
            default_version = ''
        self.tree = tree
        self.root = tree.getroot()
        self.version = self.root.get('version', default_version)
        self.author = self.root.get('author', '')
    
    def load_file(self, file_or_filename):
        etree = None
        # python2 issues
        try:
            etree = ET.parse(file_or_filename)
        except ET.ParseError:
            parser = ET.XMLParser(encoding='UTF-8')
            etree = ET.parse(file_or_filename, parser)
        self._tree_load(etree)
    
    def load_from_string(self, source, version=None):
        tree = ET.ElementTree(ET.fromstring(source))
        self._tree_load(tree, version)
    
    def get_xmlnode(self, identifier):
        xpath = ".//object[@id='{0}']".format(identifier)
        node = self.tree.find(xpath)
        return node
        
    def add_xmlnode(self, node, parent=None):
        if parent is None:
            self.root.append(node)
        else:
            parent.append(node)
        return node
    
    def add_xmlchild(self, parent, node):
        child = ET.Element('child')
        child.append(node)
        parent.append(child)
        
    def __str__(self):
        encoding='unicode'
        if sys.version_info < (3,0):
            encoding='utf-8'
        return ET.tostring(self.root, encoding=encoding)
    
    def __repr__(self):
        return '<UIFile xml="{0}">'.format(self.__str__())
        
    def save(self, file_or_filename):
        indent(self.root)
        self.tree.write(file_or_filename,
                        xml_declaration=True, encoding='utf-8')
    
    def get_widget(self, identifier):
        wmeta = None
        xpath = ".//object[@id='{0}']".format(identifier)
        node = self.tree.find(xpath)
        if node is not None:
            wmeta = self.xmlnode_to_widget(node)
        return wmeta
    
    def widgets(self):
        xpath = "./object"
        children = self.root.findall(xpath)
        for child in children:
            wmeta = self.xmlnode_to_widget(child)
            yield wmeta
    
    def widget_children(self, identifier):
        xpath = ".//object[@id='{0}']".format(identifier)
        node = self.tree.find(xpath)
        if node is not None:
            xpath = "./child/object"
            children = node.findall(xpath)
            for child in children:
                wmeta = self.xmlnode_to_widget(child)
                yield wmeta
    
    def replace_widget(self, identifier, rootmeta):
        xpath = ".//object[@id='{0}']".format(identifier)
        parent = self.root.find(xpath + '/..')
        target = parent.find(xpath)
        
        if parent is not None:
            # found something
            parent.remove(target)
            replacement = self.widget_to_xmlnode(rootmeta)
            xpath = "./child/object"
            children = target.findall(xpath)
            for child in children:
                self.add_xmlchild(replacement, child)
            if parent.tag == 'interface':
                parent.append(replacement)
            else:
                self.add_xmlchild(parent, replacement)
        


if __name__ == '__main__':
    ui = UIDefinition()
    ui.author = 'Module test'
    print(ui)
    
    xml='''<?xml version='1.0' encoding='utf-8'?>
<interface author="anonymous">
</interface>
'''
    ui.load_from_string(xml)
    print(ui, ui.author, ui.version)
    
    print('Iterating file:')
    fname = 'managers.ui'
    ui.load_file(fname)
    
    def print_widgets(w):
        print(w)
        for cw in ui.widget_children(w.identifier):
            print_widgets(cw)
    
    for w in ui.widgets():
        print_widgets(w)



# encoding: utf8
from __future__ import unicode_literals
try:
    import tkinter as tk
except:
    import Tkinter as tk
from pygubu.builder.builderobject import BuilderObject, register_widget
from pygubu.widgets.scrollbarhelper import ScrollbarHelper


class TTKSBHelperBO(BuilderObject):
    class_ = ScrollbarHelper
    container = True
    maxchildren = 1
    allowed_children = ('tk.Entry', 'ttk.Entry', 'tk.Text', 'tk.Canvas',
                        'tk.Listbox', 'ttk.Treeview' )
    OPTIONS_STANDARD = ('class_', 'cursor', 'takefocus', 'style')
    OPTIONS_SPECIFIC = ('borderwidth', 'relief', 'padding', 'height', 'width')
    OPTIONS_CUSTOM = ('scrolltype', 'usemousewheel')
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = ('class_', 'scrolltype')
    allow_bindings = False

    def get_child_master(self):
        return self.widget.container
    
    def add_child(self, bobject):
        cwidget = bobject.widget
        self.widget.add_child(cwidget)
    
    #
    # Code generation methods
    #
    def code_child_master(self):
        return '{0}.container'.format(self.code_identifier())
    
    def code_child_add(self, childid):
        lines = []
        line = '{0}.add_child({1})'.format(self.code_identifier(), childid)
        lines.append(line)
        return lines
    
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == 'usemousewheel':
            code_bag[pname] = '{0}'.format(tk.getboolean(value))
        else:
            super(TTKSBHelperBO, self)._code_set_property(targetid, pname, value, code_bag)


register_widget('pygubu.builder.widgets.scrollbarhelper', TTKSBHelperBO,
    'ScrollbarHelper', ('Pygubu Helpers', 'ttk'))



# encoding: utf8
from __future__ import unicode_literals
from collections import OrderedDict

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from .builderobject import *


#
# ttk widgets
#
class TTKWidgetBO(BuilderObject):
    OPTIONS_LABEL = ('text', 'textvariable', 'underline', 'image', 'compound',
                     'width')
    OPTIONS_COMPATIBILITY = ('state',)
    OPTIONS_STANDARD = ('class_', 'cursor', 'takefocus', 'style')
    OPTIONS_SPECIFIC = tuple()
    OPTIONS_CUSTOM = tuple()
    ro_properties = ('class_',)


class TTKFrame(TTKWidgetBO):
    OPTIONS_SPECIFIC = ('borderwidth', 'relief', 'padding', 'height', 'width')
    class_ = ttk.Frame
    container = True
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC

register_widget('ttk.Frame', TTKFrame, 'Frame', ('Containers', 'ttk'))


class TTKLabel(TTKWidgetBO):
    OPTIONS_STANDARD = \
        TTKWidgetBO.OPTIONS_STANDARD + \
        TTKWidgetBO.OPTIONS_LABEL + ('borderwidth',)
    OPTIONS_SPECIFIC = (
        'anchor', 'background', 'font', 'foreground', 'justify',
        'padding', 'relief', 'wraplength')
    class_ = ttk.Label
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC

register_widget('ttk.Label', TTKLabel, 'Label', ('Control & Display', 'ttk'))


class TTKButton(TTKWidgetBO):
    OPTIONS_STANDARD = (TTKWidgetBO.OPTIONS_STANDARD +
                        TTKWidgetBO.OPTIONS_LABEL +
                        TTKWidgetBO.OPTIONS_COMPATIBILITY)
    OPTIONS_SPECIFIC = ('command', 'default')
    class_ = ttk.Button
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC \
                    + TTKWidgetBO.OPTIONS_CUSTOM
    command_properties = ('command',)

register_widget('ttk.Button', TTKButton,
                'Button', ('Control & Display', 'ttk'))


class TTKCheckbutton(TTKWidgetBO):
    OPTIONS_STANDARD = (TTKWidgetBO.OPTIONS_STANDARD +
                        TTKWidgetBO.OPTIONS_LABEL +
                        TTKWidgetBO.OPTIONS_COMPATIBILITY)
    OPTIONS_SPECIFIC = ('command', 'offvalue', 'onvalue', 'variable')
    class_ = ttk.Checkbutton
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC \
                    + TTKWidgetBO.OPTIONS_CUSTOM
    command_properties = ('command',)

register_widget('ttk.Checkbutton', TTKCheckbutton,
                'Checkbutton', ('Control & Display', 'ttk'))


class TTKRadiobutton(TTKWidgetBO):
    OPTIONS_STANDARD = (TTKWidgetBO.OPTIONS_STANDARD +
                        TTKWidgetBO.OPTIONS_LABEL +
                        TTKWidgetBO.OPTIONS_COMPATIBILITY)
    OPTIONS_SPECIFIC = ('command', 'value', 'variable')
    class_ = ttk.Radiobutton
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC \
                    + TTKWidgetBO.OPTIONS_CUSTOM
    ro_properties = ('class_',)
    command_properties = ('command',)

register_widget('ttk.Radiobutton', TTKRadiobutton,
                'Radiobutton', ('Control & Display', 'ttk'))


class TTKCombobox(TTKWidgetBO):
    OPTIONS_SPECIFIC = ('exportselection', 'justify', 'height',
                        'postcommand', 'state', 'textvariable', 'values',
                        'width', 'validate', 'validatecommand',
                        'invalidcommand', 'xscrollcommand')
    class_ = ttk.Combobox
    container = False
    properties = (TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC +
                  TTKWidgetBO.OPTIONS_CUSTOM)
    command_properties = ('postcommand', 'validatecommand',
                          'invalidcommand', 'xscrollcommand')

register_widget('ttk.Combobox', TTKCombobox, 'Combobox',
                ('Control & Display', 'ttk'))


class TTKScrollbar(TTKWidgetBO):
    OPTIONS_SPECIFIC = ('command', 'orient')
    class_ = ttk.Scrollbar
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ('command',)

register_widget('ttk.Scrollbar', TTKScrollbar,
                'Scrollbar', ('Control & Display', 'ttk'))


class TTKSizegrip(TTKWidgetBO):
    class_ = ttk.Sizegrip
    container = False
    properties = (TTKWidgetBO.OPTIONS_STANDARD +
                  TTKWidgetBO.OPTIONS_SPECIFIC)

register_widget('ttk.Sizegrip', TTKSizegrip,
                'Sizegrip', ('Control & Display', 'ttk'))


class TTKEntry(TTKWidgetBO, EntryBaseBO):
    OPTIONS_STANDARD = TTKWidgetBO.OPTIONS_STANDARD + ('xscrollcommand',)
    OPTIONS_SPECIFIC = ('exportselection', 'font', 'invalidcommand',
                        'justify', 'show', 'state', 'textvariable',
                        'validate', 'validatecommand', 'width')
    OPTIONS_CUSTOM = ('text',)
    class_ = ttk.Entry
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ('validatecommand', 'invalidcommand',
                          'xscrollcommand')


register_widget('ttk.Entry', TTKEntry, 'Entry', ('Control & Display', 'ttk'))


class TTKProgressbar(TTKWidgetBO):
    OPTIONS_SPECIFIC = ('orient', 'length', 'mode', 'maximum',
                        'value', 'variable')  # 'phase' is read-only
    class_ = ttk.Progressbar
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC

register_widget('ttk.Progressbar', TTKProgressbar,
                'Progressbar', ('Control & Display', 'ttk'))


class TTKScale(TTKWidgetBO):
    OPTIONS_SPECIFIC = ('command', 'from_', 'length', 'orient',
                        'to', 'value', 'variable')
    class_ = ttk.Scale
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ('command',)

register_widget('ttk.Scale', TTKScale, 'Scale', ('Control & Display', 'ttk'))


class TTKSeparator(TTKWidgetBO):
    OPTIONS_SPECIFIC = ('orient',)
    class_ = ttk.Separator
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC

register_widget('ttk.Separator', TTKSeparator,
                'Separator', ('Control & Display', 'ttk'))


class TTKLabelframe(TTKWidgetBO):
    OPTIONS_STANDARD = TTKFrame.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = (TTKFrame.OPTIONS_SPECIFIC +
                        ('labelanchor', 'text', 'underline'))
    class_ = ttk.Labelframe
    container = True
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC

register_widget('ttk.Labelframe', TTKLabelframe,
                'Labelframe', ('Containers', 'ttk'))


class TTKPanedwindow(TTKWidgetBO, PanedWindowBO):
    OPTIONS_SPECIFIC = ('orient', 'height', 'width')
    class_ = ttk.Panedwindow
    allowed_children = ('ttk.Panedwindow.Pane',)
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    ro_properties = ('class_', 'orient')
    virtual_events = ('<<EnteredChild>>',)

register_widget('ttk.Panedwindow', TTKPanedwindow,
                'Panedwindow', ('Containers', 'ttk'))


class TTKNotebook(TTKWidgetBO):
    OPTIONS_SPECIFIC = ('height', 'padding', 'width')
    class_ = ttk.Notebook
    container = True
    allowed_children = ('ttk.Notebook.Tab',)
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    virtual_events = ('<<NotebookTabChanged>>',)

register_widget('ttk.Notebook', TTKNotebook,
                'Notebook', ('Containers', 'ttk'))


class TTKMenubuttonBO(TTKWidgetBO):
    OPTIONS_STANDARD = (TTKWidgetBO.OPTIONS_STANDARD +
                        TTKWidgetBO.OPTIONS_LABEL +
                        TTKWidgetBO.OPTIONS_COMPATIBILITY)
    OPTIONS_SPECIFIC = ('direction',)  # 'menu'
    class_ = ttk.Menubutton
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    allowed_children = ('tk.Menu',)
    maxchildren = 1

    def add_child(self, bobject):
        self.widget.configure(menu=bobject.widget)

register_widget('ttk.Menubutton', TTKMenubuttonBO,
                'Menubutton', ('Menu', 'Control & Display', 'ttk',))


class TTKTreeviewBO(TTKWidgetBO):
    OPTIONS_STANDARD = (TTKWidgetBO.OPTIONS_STANDARD +
                        ('xscrollcommand', 'yscrollcommand'))
    OPTIONS_SPECIFIC = ('height', 'padding', 'selectmode', 'show')
    class_ = ttk.Treeview
    container = False
    allowed_children = ('ttk.Treeview.Column',)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    virtual_events = ('<<TreeviewSelect>>', '<<TreeviewOpen>>',
                      '<<TreeviewClose>>')

    def __init__(self, builder, wdescr):
        super(TTKTreeviewBO, self).__init__(builder, wdescr)
        self._columns = None
        self._headings = None
        self._dcolumns = None

    def configure(self):
        super(TTKTreeviewBO, self).configure()
        self.__configure_columns()

    def __configure_columns(self):
        if self._columns:
            columns = list(self._columns.keys())
            if '#0' in columns:
                columns.remove('#0')
            displaycolumns = self._dcolumns
            self.widget.configure(columns=columns,
                                  displaycolumns=displaycolumns)
            for col in self._columns:
                self.widget.column(col, **self._columns[col])
        if self._headings:
            for col in self._headings:
                self.widget.heading(col, **self._headings[col])

    def set_column(self, col_id, attrs, visible=True):
        if self._columns is None:
            self._columns = OrderedDict()
            self._dcolumns = list()
        self._columns[col_id] = attrs
        if visible and col_id != '#0':
            self._dcolumns.append(col_id)

    def set_heading(self, col_id, attrs):
        if self._headings is None:
            self._headings = OrderedDict()
        self._headings[col_id] = attrs
    
    #
    # Code generation methods
    #
    def code_configure(self, targetid=None):
        if targetid is None:
            targetid = self.code_identifier()
        lines = []
        if self._columns:
            columns = list(self._columns.keys())
            if '#0' in columns:
                columns.remove('#0')
            displaycolumns = self._dcolumns
            line = '{0}_cols = {1}'.format(targetid, repr(columns))
            lines.append(line)
            line = '{0}_dcols = {1}'.format(targetid, repr(displaycolumns))
            lines.append(line)
            line = '{0}.configure(columns={0}_cols, displaycolumns={0}_dcols)'
            line = line.format(targetid)
            lines.append(line)
            for col in self._columns:
                code_bag, kwp, _ = self._code_process_properties(
                    self._columns[col], targetid)
                bag = []
                for pname in kwp:
                    s = '{0}={1}'.format(pname, code_bag[pname])
                    bag.append(s)
                kwargs = ','.join(bag)
                line = "{0}.column('{1}', {2})".format(targetid, col, kwargs)
                lines.append(line)
        if self._headings:
            for col in self._headings:
                code_bag, kwp, _ = self._code_process_properties(
                    self._headings[col], targetid)
                bag = []
                for pname in kwp:
                    s = '{0}={1}'.format(pname, code_bag[pname])
                    bag.append(s)
                kwargs = ','.join(bag)
                line = "{0}.heading('{1}', {2})".format(targetid, col, kwargs)
                lines.append(line)
        return lines


register_widget('ttk.Treeview', TTKTreeviewBO,
                'Treeview', ('Control & Display', 'ttk'))


#
# Helpers for Standard ttk widgets
#

class TTKPanedwindowPane(TTKWidgetBO, PanedWindowPaneBO):
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = ('weight',)
    class_ = None
    container = True
    allowed_parents = ('ttk.Panedwindow',)
    maxchildren = 1
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC

register_widget('ttk.Panedwindow.Pane', TTKPanedwindowPane,
                'Panedwindow.Pane', ('Pygubu Helpers', 'ttk'))


class TTKNotebookTab(TTKWidgetBO):
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = ('state', 'sticky', 'padding', 'text',
                        'image', 'compound', 'underline')
    class_ = None
    container = True
    layout_required = False
    allow_bindings = False
    allowed_parents = ('ttk.Notebook',)
    maxchildren = 1
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC

    def realize(self, parent):
        self.widget = parent.get_child_master()
        return self.widget

    def configure(self):
        pass

    def layout(self):
        pass

    def add_child(self, bobject):
        self.widget.add(bobject.widget, **self.wmeta.properties)
        
    #
    # Code generation methods
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        return tuple()
    
    def code_configure(self, targetid=None):
        return tuple()
    
    def code_child_add(self, childid):
        targetid = self.code_identifier()
        code_bag, kw, _ = self._code_process_properties(self.wmeta.properties,
                                                 targetid)
        kwbag = []
        for pname in kw:
            arg = '{0}={1}'.format(pname, code_bag[pname])
            kwbag.append(arg)
        kwargs = ''
        if kwbag:
            kwargs = ', {0}'.format(', '.join(kwbag))
        line = '{0}.add({1}{2})'.format(targetid, childid, kwargs)
        return [line]


register_widget('ttk.Notebook.Tab', TTKNotebookTab,
                'Notebook.Tab', ('Pygubu Helpers', 'ttk'))


class TTKTreeviewColBO(TTKWidgetBO):
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = ('text', 'image', 'command', 'heading_anchor',
                        'column_anchor', 'minwidth', 'stretch', 'width')
    OPTIONS_CUSTOM = ('tree_column', 'visible',)
    class_ = None
    container = False
    layout_required = False
    allow_bindings = False
    allowed_parents = ('ttk.Treeview',)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ('command',)

    def realize(self, parent):
        self.widget = parent.get_child_master()
        col_props = dict(self.wmeta.properties)  # copy properties
        self._setup_column(parent, col_props)
        return self.widget
    
    def _setup_column(self, parent, col_props):
        tree_column = col_props.pop('tree_column', 'false')
        tree_column = tree_column.lower()
        tree_column = True if tree_column == 'true' else False
        column_id = '#0' if tree_column else self.wmeta.identifier
        visible = col_props.pop('visible', 'false')
        visible = visible.lower()
        is_visible = True if visible == 'true' else False

        #configure heading properties
        col_props.pop('command', '')
        hprops = {
            'anchor': col_props.pop('heading_anchor', tk.W),
            'text': col_props.pop('text', '')
        }
        # Only add image if has value. Fix code generation
        imgvalue = col_props.pop('image', None)
        if imgvalue:
            hprops['image'] = imgvalue
        
        parent.set_heading(column_id, hprops)

        #configure column properties
        cprops = {
            'anchor': col_props.pop('column_anchor', ''),
            'stretch': col_props.pop('stretch', '1'),
            'width': col_props.pop('width', '200'),
            'minwidth': col_props.pop('minwidth', '20')
        }
        parent.set_column(column_id, cprops, is_visible)

    def configure(self):
        pass

    def layout(self):
        pass

    def _connect_command(self, cpname, callback):
        tree_column = self.wmeta.properties.get('tree_column', 'false')
        tree_column = tree_column.lower()
        tree_column = True if tree_column == 'true' else False
        column_id = '#0' if tree_column else self.wmeta.identifier
        self.widget.heading(column_id, command=callback)
    
    #
    # Code generation methods
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        col_props = dict(self.wmeta.properties)  # copy properties
        self._setup_column(boparent, col_props)
        return tuple()
    
    def code_configure(self, targetid=None):
        return tuple()


register_widget('ttk.Treeview.Column', TTKTreeviewColBO,
                'Treeview.Column', ('Pygubu Helpers', 'ttk'))


class TTKSpinboxBO(TTKWidgetBO, EntryBaseBO):
    OPTIONS_STANDARD = TTKEntry.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = (TTKEntry.OPTIONS_SPECIFIC +
                        ('from_', 'to', 'increment', 'values',
                         'wrap', 'format', 'command'))
    OPTIONS_CUSTOM = TTKEntry.OPTIONS_CUSTOM
    class_ = None
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ('validatecommand', 'invalidcommand',
                          'xscrollcommand', 'command')
    virtual_events = ('<<Increment>>', '<<Decrement>>')


if tk.TkVersion >= 8.6:
    if not hasattr(ttk, 'Spinbox'):
        from pygubu.widgets.ttkspinbox import Spinbox
        ttk.Spinbox = Spinbox

    TTKSpinboxBO.class_ = ttk.Spinbox

    register_widget('ttk.Spinbox', TTKSpinboxBO,
                    'Spinbox', ('Control & Display', 'ttk')
                    )



#!/usr/bin/env python3
# encoding: UTF-8

"""Build tar.gz for pygubu

Needed packages to run (using Debian/Ubuntu package names):

    python3-tk
"""
import pygubu

try:
    from setuptools import setup
except:
    from distutils.core import setup

VERSION = pygubu.__version__

long_description = \
"""
Welcome to pygubu a GUI designer for tkinter
============================================

Pygubu is a RAD tool to enable quick & easy development of user interfaces
for the python tkinter module.

The user interfaces designed are saved as XML, and by using the pygubu builder
these can be loaded by applications dynamically as needed.
Pygubu is inspired by Glade.

IMPORTANT. Since version 0.10 pygubu was splited in two projects, the core library (this package)
and the gui designer (pygubu-designer).


Installation
------------

Pygubu requires python >= 2.7 (Tested only in python 2.7.3 and 3.2.3
with tk8.5)

Download and extract the tarball. Open a console in the extraction
path and execute:

::

    python setup.py install


Usage
-----

Create an UI definition using pygubu-designer and save it to a file. Then, 
create your aplication script as shown below. Note that 'mainwindow' 
is the name of your Toplevel widget.

::

    # helloworld.py
    import tkinter as tk
    import pygubu
    
    
    class HelloWorldApp:
        
        def __init__(self):
    
            #1: Create a builder
            self.builder = builder = pygubu.Builder()
    
            #2: Load an ui file
            builder.add_from_file('helloworld.ui')
    
            #3: Create the mainwindow
            self.mainwindow = builder.get_object('mainwindow')
            
        def run(self):
            self.mainwindow.mainloop()


    if __name__ == '__main__':
        app = HelloWorldApp()
        app.run()


See the examples directory or watch this hello world example on
video http://youtu.be/wuzV9P8geDg
"""

setup(
    name='pygubu',
    version=VERSION,
    license='MIT',
    author='Alejandro Autalán',
    author_email='alejandroautalan@gmail.com',
    description='A tkinter GUI builder.',
    long_description=long_description,
    url='https://github.com/alejandroautalan/pygubu',

    packages=['pygubu', 'pygubu.builder',
              'pygubu.builder.widgets', 'pygubu.widgets'],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Topic :: Software Development :: User Interfaces",
    ],
)



# encoding: utf8
from __future__ import unicode_literals

try:
    import tkinter as tk
    import tkinter.ttk as ttk
    from tkinter import filedialog
except:
    import Tkinter as tk
    import ttk
    import tkFileDialog as filedialog


class PathChooserInput(ttk.Frame):
    """ Allows to choose a file or directory.
    
    Generates <<PathChooserPathChanged>> event when the path is changed.
    
    Dialog options:
      initialdir: str
      filetypes: iterable
      title: str
      mustexist: str
    
    Usage Example:
        # Choose File:
        pcifile = PathChooserInput(framex)
        pcifile.config(initialdir='/home', title='Choose a file:', type='file')
        pcifile.config(filetypes=[('text files-', '.txt'), ('uifiles', '.ui')])
        pcifile.pack(fill='x', side='top')
        
        # Choose directory:
        pcidir = PathChooserInput(framex)
        pcidir.config(initialdir='/usr/local', mustexist='true',
                      title='Choose a directory:', type='directory')
        pcidir.pack(fill='x', side='top')
    """

    FILE = 'file'
    DIR = 'directory'

    def __init__(self, master=None, **kw):
        ttk.Frame.__init__(self, master, **kw)
        self._choose = self.FILE
        self._oldvalue = ''
        self._state = 'normal'
        self._fdoptions = {
            'filetypes': tuple(),
            'initialdir': None,
            'mustexist': False,
            'title': None,
            }
        # subwidgets
        self.entry = o = ttk.Entry(self, state=self._state)
        o.grid(row=0, column=0, sticky='ew')
        o.bind('<KeyPress>', self.__on_enter_key_pressed)
        o.bind('<FocusOut>', self.__on_focus_out)
        self.folder_button = o = ttk.Button(
                self,
                text='▶',
                command=self.__on_folder_btn_pressed,
                width=4,
                state=self._state)
        o.grid(row=0, column=1, padx=2)

        #self.rowconfigure(0, weight = 0)
        self.columnconfigure(0, weight = 1)

    def configure(self, cnf=None, **kw):
        args = tk._cnfmerge((cnf, kw))
        key = 'type'
        if key in args:
            self._choose = args[key]
            del args[key]
        key = 'image'
        if key in args:
            self.folder_button.configure(image=args[key])
            del args[key]
        key = 'path'
        if key in args:
            self.entry.delete(0, 'end')
            self.entry.insert(0, args[key])
            self._generate_changed_event()
            del args[key]
        key = 'textvariable'
        if key in args:
            self.entry.configure(textvariable=args[key])
            self._generate_changed_event()
            del args[key]
        key = 'state'
        if key in args:
            value = args[key]
            self.entry.config(state=value)
            if value in ('disabled', 'readonly'):
                self.folder_button.config(state='disabled')
            else:
                self.folder_button.config(state=value)
            del args[key]
        # dialog options
        for key in tuple(args.keys()):
            if key in self._fdoptions:
                self._fdoptions[key] = args[key]
                args.pop(key)
        ttk.Frame.configure(self, args)

    config = configure

    def cget(self, key):
        option = 'type'
        if key == option:
            return self._choose
        option = 'image'
        if key == option:
            return self.folder_button.cget(key)
        option = 'path'
        if key == option:
            return self.entry.get()
        option = 'textvariable'
        if key == option:
            return self.entry.cget(option)
        option = 'state'
        if key == option:
            return self.entry.cget(option)
        # dialog options
        if key in self._fdoptions:
            return self._fdoptions[key]
        return ttk.Frame.cget(self, key)

    __getitem__ = cget
        
    def _is_changed(self):
#        print(repr(self._oldvalue), ':', repr(self.entry.get()))
        if self._oldvalue != self.entry.get():
            return True
        return False
    
    def _generate_changed_event(self):
        if self._is_changed():
            self._oldvalue = self.entry.get()
            self.event_generate('<<PathChooserPathChanged>>')

    def __on_enter_key_pressed(self, event):
        key = event.keysym
        if key in  ('Return','KP_Enter'):
            self._generate_changed_event()
    
    def __on_focus_out(self, event):
         self._generate_changed_event()

    def __on_folder_btn_pressed(self):
        fname = None
        fdoptions = self._fdoptions.copy()
        if fdoptions['initialdir'] is None:
            fdoptions['initialdir'] = self.cget('path')
        if self._choose == self.FILE:
            fdoptions.pop('mustexist')
            fname = filedialog.askopenfilename(**fdoptions)
        else:
            fdoptions.pop('filetypes')
            fname = filedialog.askdirectory(**fdoptions)
        if fname:
            self.configure(path=fname)




# encoding: utf8
from __future__ import unicode_literals

try:
    import tkinter as tk
except:
    import Tkinter as tk

from pygubu.builder.builderobject import BuilderObject, register_widget
from pygubu.widgets.tkscrolledframe import TkScrolledFrame


class TKScrolledFrameBO(BuilderObject):
    class_ = TkScrolledFrame
    container = True
#    maxchildren = 1
#    allowed_children = ('tk.Frame', 'ttk.Frame' )
    OPTIONS_STANDARD = ('borderwidth', 'cursor', 'highlightbackground',
                        'highlightcolor', 'highlightthickness',
                        'padx', 'pady', 'relief', 'takefocus')
    OPTIONS_SPECIFIC = ('background',  'class_', 'container',
                        'height', 'width')
    OPTIONS_CUSTOM = ('scrolltype', 'usemousewheel')
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = ('class_', 'scrolltype')

    def get_child_master(self):
        return self.widget.innerframe

    def configure(self, target=None):
        super(TKScrolledFrameBO, self).configure(self.widget.innerframe)
        
    def _set_property(self, target_widget, pname, value):
        if pname in ('usemousewheel',):
            super(TKScrolledFrameBO, self)._set_property(self.widget, pname, value)
        else:
            super(TKScrolledFrameBO, self)._set_property(target_widget, pname, value)

    #def layout(self, target=None, configure_gridrc=True):
    #    super(TKScrolledFrameBO, self).layout(target, False)
    #    self._gridrc_config(self.widget.innerframe)
    
    #
    # Code generation methods
    #
    def code_child_master(self):
        return '{0}.innerframe'.format(self.code_identifier())
    
    def code_configure(self, targetid=None):
        realtarget = '{0}.innerframe'.format(self.code_identifier())
        return super(TKScrolledFrameBO,self).code_configure(realtarget)
    
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == 'usemousewheel':
            nvalue = '{0}.configure({1}={2})'.format(self.code_identifier(),
                                                     pname,
                                                     tk.getboolean(value))
            code_bag[pname] = [nvalue]
        else:
            super(TKScrolledFrameBO, self)._code_set_property(targetid, pname,
                                                             value, code_bag)

register_widget('pygubu.builder.widgets.tkscrolledframe', TKScrolledFrameBO,
    'ScrolledFrame', ('Pygubu Widgets', 'tk'))
